# Copyright(C) [2025] Advanced Micro Devices, Inc. All rights reserved.
"""
Testing the (FP8) case of a dot op that consumes the output (MFMA) of
another dot op as an input.

"""
######################################## Imports#######################################
import math
import pytest
import torch

import triton
import triton.language as tl
######################################## Imports#######################################

########################## HELPER utils ##########################
TORCH_HAS_FP8E4 = hasattr(torch, 'float8_e4m3fnuz')
float8: tl.constexpr = None if not TORCH_HAS_FP8E4 else torch.float8_e4m3fnuz
########################## HELPER utils ##########################


@triton.jit
def _chained_dot(
    Q,
    K,
    V,
    Out,
    q_desc,
    k_desc,
    v_desc,
    s_sc,
    s_desc,
    o_sc,
    stride_qz,
    stride_qm,
    stride_qd,
    stride_kz,
    stride_kn,
    stride_kd,
    stride_vz,
    stride_vd,
    stride_vn,
    stride_oz,
    stride_om,
    stride_od,
    Z,
    M,
    N,
    BLOCK_D: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    USE_FP8: tl.constexpr,
):
    start_m = tl.program_id(0)
    off_z = tl.program_id(1)
    qkv_offset = off_z * stride_qz
    Q_block_ptr = tl.make_block_ptr(base=Q + qkv_offset, shape=(N, BLOCK_D), strides=(stride_qm, stride_qd),
                                    offsets=(start_m * BLOCK_M, 0), block_shape=(BLOCK_M, BLOCK_D), order=(1, 0))
    K_block_ptr = tl.make_block_ptr(base=K + qkv_offset, shape=(BLOCK_D, N), strides=(stride_kd, stride_kn),
                                    offsets=(0, 0), block_shape=(BLOCK_D, BLOCK_N), order=(0, 1))
    V_block_ptr = tl.make_block_ptr(base=V + qkv_offset, shape=(N, BLOCK_D), strides=(stride_vn, stride_vd),
                                    offsets=(0, 0), block_shape=(BLOCK_N, BLOCK_D), order=(0, 1))

    s_scale = q_desc * k_desc * s_sc
    acc_scale = s_desc * v_desc * o_sc

    q = tl.load(Q_block_ptr)

    acc = tl.zeros([BLOCK_M, BLOCK_D], dtype=tl.float32)
    lo, hi = 0, N
    for start_n in range(lo, hi, BLOCK_N):
        start_n = tl.multiple_of(start_n, BLOCK_N)

        k = tl.load(K_block_ptr)
        s = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
        s += tl.dot(q, k)

        if USE_FP8:
            s *= s_scale

        v = tl.load(V_block_ptr)
        acc += tl.dot(s.to(v.dtype), v)

        K_block_ptr = tl.advance(K_block_ptr, (0, BLOCK_N))
        V_block_ptr = tl.advance(V_block_ptr, (BLOCK_N, 0))

    if USE_FP8:
        acc *= acc_scale

    O_block_ptr = tl.make_block_ptr(base=Out + qkv_offset, shape=(N, BLOCK_D), strides=(stride_om, stride_od),
                                    offsets=(start_m * BLOCK_M, 0), block_shape=(BLOCK_M, BLOCK_D), order=(1, 0))
    tl.store(O_block_ptr, acc.to(Out.type.element_ty))

##################################################################################################################################################  

import numpy as np
import random
import torch 
import os
from geak_eval.perf.ROCm.performance_utils_pytest import PytestBenchmarker, do_bench_config, save_all_benchmark_results
from typing import Dict
########################## HELPER utils ##########################
TORCH_HAS_FP8E4 = hasattr(torch, 'float8_e4m3fnuz')
float8: tl.constexpr = None if not TORCH_HAS_FP8E4 else torch.float8_e4m3fnuz
########################## HELPER utils ##########################


result_gold = {}
######################################## HELPERS for Eval ######################################## 
def set_seed(seed: int = 42) -> None:
    """
    Set the random seed for reproducibility across multiple libraries and configure PyTorch for deterministic behavior.

    Args:
        seed (int): The seed value to set. Default is 42.
    """
    # Set seed for Python's built-in random module
    random.seed(seed)
    # Set seed for NumPy
    np.random.seed(seed)
    # Set seed for PyTorch on CPU
    torch.manual_seed(seed)
    # Set seed for PyTorch on all GPUs (if available)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    # Ensure deterministic behavior in PyTorch
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # Set environment variable for hash-based operations
    os.environ['PYTHONHASHSEED'] = str(seed)

######################################## HELPERS for Eval ######################################## 
class chained_dot_fn(torch.autograd.Function):

    @staticmethod
    def forward(ctx, q, k, v, msize=32, q_desc=1.0, k_desc=1.0, v_desc=1.0, s_sc=1.0, s_desc=1.0, o_sc=1.0):
        Lq, Lk, Lv = q.shape[-1], k.shape[-1], v.shape[-2]
        assert Lq == Lk and Lk == Lv
        assert Lk in {16, 32, 64, 128}
        assert msize in {16, 32}
        o = torch.empty_like(q, dtype=v.dtype)

        BLOCK_M = 128 if q.dtype == float8 else 256
        if BLOCK_M > q.shape[1]:
            BLOCK_M = int(math.pow(2, math.floor(math.log2(q.shape[1]))))
        BLOCK_N = 32
        if BLOCK_N > k.shape[1]:
            BLOCK_N = int(math.pow(2, math.floor(math.log2(k.shape[1]))))
        waves_per_eu = 2
        num_warps = 4 if q.dtype == float8 else 8
        num_stages = 1

        grid = (triton.cdiv(q.shape[1], BLOCK_M), q.shape[0], 1)

        _chained_dot[grid](q, k, v, o, q_desc,
                           k_desc, v_desc, s_sc, s_desc, o_sc, q.stride(0), q.stride(1), q.stride(2), k.stride(0),
                           k.stride(1), k.stride(2), v.stride(0), v.stride(1), v.stride(2), o.stride(0), o.stride(1),
                           o.stride(2), Z=q.shape[0], M=q.shape[1], N=k.shape[1], BLOCK_D=Lk, BLOCK_M=BLOCK_M,
                           BLOCK_N=BLOCK_N, USE_FP8=(q.dtype == float8), waves_per_eu=waves_per_eu, num_warps=num_warps,
                           num_stages=num_stages, matrix_instr_nonkdim=msize)

        return o


chained_dot = chained_dot_fn.apply


def to_float8(x, dtype=float8, margin: float = 1.0):
    finfo = torch.finfo(dtype)
    scale = finfo.max / x.abs().max().clamp(min=1e-12)
    scale = math.pow(2, math.floor(math.log2(scale.float().item())) - margin)
    x_scaled = (x.float() * scale).clamp(min=finfo.min, max=finfo.max)
    return x_scaled.to(dtype), scale, 1.0 / scale


@pytest.mark.parametrize('M, N, D, dtype, msize', [(*shape, dtype, msize)
                                                   for shape in [(128, 64, 32), (256, 128, 128)]
                                                   for dtype in ['fp8']
                                                   for msize in [16, 32]])
def test_chained_dot(M, N, D, dtype, msize,request):
    set_seed()
    if dtype == 'fp8':
        assert float8 is not None

    BATCH = 1
    q = torch.empty((BATCH, M, D), dtype=torch.float16, device="cuda").normal_(mean=0., std=0.5)
    k = torch.empty((BATCH, N, D), dtype=torch.float16, device="cuda").normal_(mean=0., std=0.5)
    v = torch.empty((BATCH, D, N), dtype=torch.float16, device="cuda").normal_(mean=0., std=0.5)

    if dtype == 'fp8':
        q_f8, _, q_desc = to_float8(q)
        k_f8, _, k_desc = to_float8(k)
        v_f8, _, v_desc = to_float8(v)

        s = torch._scaled_mm(q_f8[0], k_f8[0].transpose(0, 1), out_dtype=torch.float32,
                             scale_a=torch.tensor(q_desc, dtype=torch.float32, device="cuda"),
                             scale_b=torch.tensor(k_desc, dtype=torch.float32, device="cuda"))
        s_f8, s_sc, s_desc = to_float8(s)
        ref = torch._scaled_mm(s_f8, v_f8[0].transpose(0, 1), out_dtype=torch.float32,
                               scale_a=torch.tensor(s_desc, dtype=torch.float32, device="cuda"),
                               scale_b=torch.tensor(v_desc, dtype=torch.float32, device="cuda"))
        ref_f8, ref_sc, _ = to_float8(ref)

        tri_out = chained_dot(q_f8, k_f8, v_f8, msize, q_desc, k_desc, v_desc, s_sc, s_desc, ref_sc)
        result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])
        ################### save tri_out in result_gold ###################
        test_case_name = request.node.name
        sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_") + "_fwd"
        result_gold[sanitized_key_name] = tri_out[0].float().clone().detach().cpu()
        ###################################################################


        assert tri_out.isnan().sum() == 0
        torch.testing.assert_close(tri_out[0].float(), ref_f8.float(), atol=1e-2, rtol=0)

    else:
        s = torch.matmul(q, k.transpose(1, 2))
        ref = torch.matmul(s, v.transpose(1, 2))

        tri_out = chained_dot(q, k, v, msize)
        result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])
        ################### save tri_out in result_gold ###################
        test_case_name = request.node.name
        sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_")
        result_gold[sanitized_key_name] = tri_out.clone().detach().cpu()
        ###################################################################

        torch.testing.assert_close(tri_out, ref, atol=1e-2, rtol=0)

# --- Define TFLOPS and GB/s calculators ---
def calculate_chained_dot_tflops(params: dict, ms: float) -> float:
    BATCH, M, N_kv, D = params['BATCH'], params['M_seqlen'], params['N_kv_seqlen'], params['D_head']
    # S = Q(M,D) @ K.T(D,N_kv) -> 2*M*N_kv*D
    # O = S(M,N_kv) @ V.T(N_kv,D) -> 2*M*N_kv*D (assuming V value dim is D)
    flops = BATCH * ( (2 * M * N_kv * D) + (2 * M * N_kv * D) )
    tflops = flops / (ms / 1000) / 1e12
    return tflops

def calculate_chained_dot_gbps(params: dict, ms: float) -> float:
    BATCH, M, N_kv, D = params['BATCH'], params['M_seqlen'], params['N_kv_seqlen'], params['D_head']
    dtype_str = params.get('dtype_str', 'fp16')
    
    element_size = 1 if dtype_str == 'fp8' and TORCH_HAS_FP8E4 and float8 is not None else \
                   (4 if dtype_str == 'fp32' else 2) # Default to 2 bytes for fp16/bf16

    bytes_q = BATCH * M * D * element_size
    bytes_k = BATCH * N_kv * D * element_size
    # V is passed as (B,D,N_kv) to kernel, effectively same number of elements as K
    bytes_v = BATCH * D * N_kv * element_size 
    bytes_o = BATCH * M * D * element_size # Output O is (B,M,D)
    total_bytes = bytes_q + bytes_k + bytes_v + bytes_o
    gbps = total_bytes / (ms / 1000) / 1e9
    return gbps

OP_NAME_FOR_BENCHMARK = "chained_dot_fp8_perf"

# --- Pytest parametrize for performance testing ---
CHAINED_DOT_PERF_CONFIGS = []
# M, N_kv, D_head
base_shapes_perf = [(128, 64, 32), (256, 128, 64), (512, 128, 128), (1024, 256, 128), (2048, 512, 64)]
dtypes_perf_list = ['fp16']
if TORCH_HAS_FP8E4 and float8 is not None: # Check if the global float8 (torch.dtype or None) is set
    dtypes_perf_list.append('fp8')
msize_args_perf_list = [16, 32] # msize from original test

for shape_cfg in base_shapes_perf:
    m_val, n_kv_val, d_val = shape_cfg
    for dtype_s_val in dtypes_perf_list:
        for msize_val_val in msize_args_perf_list:
            CHAINED_DOT_PERF_CONFIGS.append({
                'M':m_val, 'N_kv':n_kv_val, 'D':d_val, 
                'dtype_str':dtype_s_val, 'msize':msize_val_val
            })

@pytest.mark.parametrize("test_config", CHAINED_DOT_PERF_CONFIGS)
def test_performance(test_config, request):
    set_seed()
    M_seqlen = test_config['M']
    N_kv_seqlen = test_config['N_kv']
    D_head = test_config['D']
    dtype_str = test_config['dtype_str']
    msize_arg = test_config['msize']

    BATCH = 2 # Example batch size for performance test

    # Prepare inputs (always start with fp16 for data generation simplicity before potential cast)
    q_host = torch.empty((BATCH, M_seqlen, D_head), dtype=torch.float16, device="cuda").normal_(mean=0., std=0.5)
    k_host = torch.empty((BATCH, N_kv_seqlen, D_head), dtype=torch.float16, device="cuda").normal_(mean=0., std=0.5)
    # V is passed as (B, D, N_kv) to chained_dot call in original test
    v_host_for_call = torch.empty((BATCH, D_head, N_kv_seqlen), dtype=torch.float16, device="cuda").normal_(mean=0., std=0.5)

    q_for_kernel, k_for_kernel, v_for_kernel_call = q_host, k_host, v_host_for_call
    # Python floats for scales, matching chained_dot_fn.forward signature
    q_desc_py, k_desc_py, v_desc_py = 1.0, 1.0, 1.0
    s_sc_py, s_desc_py, o_sc_py = 1.0, 1.0, 1.0

    if dtype_str == 'fp8':
        if not (TORCH_HAS_FP8E4 and float8 is not None): # Check global float8
            pytest.skip("FP8 (e4m3fnuz) not available or global float8 is None.")
        
        # to_float8's default dtype is the global `float8` (torch.dtype or None)
        q_f8, _, q_desc_float = to_float8(q_host) # Returns Python float for scales
        k_f8, _, k_desc_float = to_float8(k_host)
        v_f8, _, v_desc_float = to_float8(v_host_for_call)
        
        q_for_kernel, k_for_kernel, v_for_kernel_call = q_f8, k_f8, v_f8
        q_desc_py, k_desc_py, v_desc_py = q_desc_float, k_desc_float, v_desc_float

        # Derive s_sc_py, s_desc_py, o_sc_py using the reference path from original test
        # Use BATCH=0 for these calculations as original test did
        s_ref_fp32 = torch._scaled_mm(
            q_f8[0], k_f8[0].transpose(0, 1), 
            out_dtype=torch.float32,
            scale_a=torch.tensor(q_desc_py, dtype=torch.float32, device="cuda"), # _scaled_mm needs tensor scales
            scale_b=torch.tensor(k_desc_py, dtype=torch.float32, device="cuda")
        )
        _s_f8_ref, s_sc_float, s_desc_float = to_float8(s_ref_fp32)
        
        # v_f8[0] is (D, N_kv). Transpose to (N_kv, D) for S(M,N_kv) @ V.T(N_kv,D)
        o_ref_fp32 = torch._scaled_mm(
            _s_f8_ref, v_f8[0].transpose(0, 1), 
            out_dtype=torch.float32, 
            scale_a=torch.tensor(s_desc_float, dtype=torch.float32, device="cuda"), 
            scale_b=torch.tensor(v_desc_float, dtype=torch.float32, device="cuda")
        )
        _o_f8_ref, o_sc_float, _ = to_float8(o_ref_fp32)

        s_sc_py, s_desc_py, o_sc_py = s_sc_float, s_desc_float, o_sc_float
    
    # --- Create op_lambda for benchmarking ---
    op_lambda = lambda: chained_dot( # This is chained_dot_fn.apply
        q_for_kernel, k_for_kernel, v_for_kernel_call, msize_arg,
        q_desc_py, k_desc_py, v_desc_py,
        s_sc_py, s_desc_py, o_sc_py
    )

    # --- Benchmarking ---
    bench_config = do_bench_config(warm_up=10, repetition=50) # Adjust reps as needed
    benchmarker = PytestBenchmarker(op_callable=op_lambda,
                                    op_name=OP_NAME_FOR_BENCHMARK,
                                    config=bench_config)

    current_params_for_logs_and_calc = {
        "BATCH": BATCH, "M_seqlen": M_seqlen, "N_kv_seqlen": N_kv_seqlen, "D_head": D_head,
        "dtype_str": dtype_str, "msize_arg": msize_arg
    }
    benchmarker.run_benchmark(current_params_dict=current_params_for_logs_and_calc,
                              gbps_calculator=calculate_chained_dot_gbps,
                              tflops_calculator=calculate_chained_dot_tflops)
    
######################################## HELPERS for Eval ########################################     
# --- Pytest hook to save the dictionary at the end of the session ---  
def test_save_results():  
    """  
    Called after whole test run finished, right before returning the exit status to the system.  
    """
    print('Inside session finish...')
    if "_CALL_SUCCESS_" not in result_gold:
        result_gold['_CALL_SUCCESS_'] = torch.tensor([[0.0]])
    OUTPUT_FILENAME = __file__.replace('.','_') + '.pt'
    print(f"\nSaving all y_triton results to {OUTPUT_FILENAME}...")  
    # Ensure the directory for the output file exists if it's in a subdirectory  
    output_dir = os.path.dirname(OUTPUT_FILENAME)  
    if output_dir and not os.path.exists(output_dir):  
        os.makedirs(output_dir, exist_ok=True)  
    torch.save(result_gold, OUTPUT_FILENAME)       
    print(f"Successfully saved {len(result_gold)} y_triton tensors to {OUTPUT_FILENAME}.") 


def test_save_performance_results():
    """
    Called after the test_performance function finishes.
    This is a separate hook to ensure performance results are saved.
    """
    print('\nPytest session finishing... Saving benchmark results...')

    output_directory = os.path.join(os.path.dirname(__file__), "perf")  # Save in a "perf" subdirectory next to the test file
    os.makedirs(output_directory, exist_ok=True)
    
    save_all_benchmark_results(output_directory)
    print(f"All benchmark results attempted to save to: {output_directory}")

######################################## HELPERS for Eval ######################################## 


# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def _fp4_packed_to_bf16(
    x_packed,
    sign_mask_f4,
    mantissa_mask_f4,
    mbits_f4_e2m1,
    ebits_f4_e2m1,
    f4_e2m1_exp_bias,
    mbits_f32,
    ebits_f32,
    f32_exp_bias,
    zero_bits_f32,
    zero_point_five_bits_f32,
):
    """
    Input: a tensor of packed fp4 values
    Output: a tensor of bfloat16 values
    """

    # low-bits: original location 0:3
    # high-bits: original location 4:7
    x_low_bits = x_packed >> 4
    x_high_bits = x_packed & 0xF
    x = tl.interleave(x_low_bits, x_high_bits)

    # cast logic below
    # output = x_unpacked.to(tl.float32)

    # save the sign
    sign_f4 = x & sign_mask_f4

    # set everything to positive, will add sign back at the end
    x_pos = x ^ sign_f4

    # Special case zero
    zero_mask = x_pos == 0

    # There is only one denormal value in fp4: s001, which is 0.5 in f32
    # Special case it.
    # TODO(later): will it be faster to repeat this for all 8 positive
    # values instead of the bit manipulations?
    denormal_mask = x_pos == 1

    # calculate the new exponent and shift it to bits 2:9 of the result
    exp_biased_f4 = x_pos >> mbits_f4_e2m1
    exp_biased_f32 = exp_biased_f4 - f4_e2m1_exp_bias + f32_exp_bias
    exp_biased_f32 = exp_biased_f32.to(tl.int32) << mbits_f32

    # shift the mantissa to bits 10:32 of the result
    mantissa_f4 = x_pos & mantissa_mask_f4
    mantissa_f32 = mantissa_f4.to(tl.int32) << (mbits_f32 - mbits_f4_e2m1)
    output = mantissa_f32

    # combine the pieces
    result = exp_biased_f32 | mantissa_f32
    # result[zero_mask] = ZERO_BITS_F32
    result = tl.where(zero_mask, zero_bits_f32, result)
    # result[denormal_mask] = ZERO_POINT_FIVE_BITS_F32
    result = tl.where(denormal_mask, zero_point_five_bits_f32, result)

    # add sign back
    sign_f32 = sign_f4.to(tl.int32) << (
        mbits_f32 - mbits_f4_e2m1 + ebits_f32 - ebits_f4_e2m1
    )
    result = result | sign_f32

    # The bit shifting above is for float32, so for now we
    # bitcast to float32 and then regular cast to bfloat16
    # TODO(later): it should be pretty easy to cast directly to bf16, just
    # need to adjust the mbits/ebits/special values. Perf impact is likely
    # to be small as we would not be chaning memory access patterns.
    output = result.to(tl.float32, bitcast=True)
    output = output.to(tl.bfloat16)
    return output

@triton.jit
def triton_f4_to_bf16_kernel(
    x_ptr,
    output_ptr,
    n_elements_in,
    sign_mask_f4: tl.constexpr,
    mantissa_mask_f4: tl.constexpr,
    mbits_f4_e2m1: tl.constexpr,
    ebits_f4_e2m1: tl.constexpr,
    f4_e2m1_exp_bias: tl.constexpr,
    mbits_f32: tl.constexpr,
    ebits_f32: tl.constexpr,
    f32_exp_bias: tl.constexpr,
    zero_bits_f32: tl.constexpr,
    zero_point_five_bits_f32: tl.constexpr,
    BLOCK_SIZE_IN: tl.constexpr,
):
    pid = tl.program_id(axis=0)
    n_elements_out = n_elements_in * 2
    BLOCK_SIZE_OUT: tl.constexpr = BLOCK_SIZE_IN * 2

    block_start_in = pid * BLOCK_SIZE_IN
    offsets_in = block_start_in + tl.arange(0, BLOCK_SIZE_IN)

    mask_in = offsets_in < n_elements_in

    # packed uint8
    x_packed = tl.load(x_ptr + offsets_in, mask=mask_in)
    output = _fp4_packed_to_bf16(
        x_packed,
        sign_mask_f4,
        mantissa_mask_f4,
        mbits_f4_e2m1,
        ebits_f4_e2m1,
        f4_e2m1_exp_bias,
        mbits_f32,
        ebits_f32,
        f32_exp_bias,
        zero_bits_f32,
        zero_point_five_bits_f32,
    )

    # set up output offsets
    block_start_out = pid * BLOCK_SIZE_OUT
    offsets_out = block_start_out + tl.arange(0, BLOCK_SIZE_OUT)
    mask_out = offsets_out < n_elements_out

    tl.store(output_ptr + offsets_out, output, mask=mask_out)

EBITS_F32, MBITS_F32 = 8, 23
EBITS_F4_E2M1, MBITS_F4_E2M1 = 2, 1
EBITS_F6_E2M3, MBITS_F6_E2M3 = 2, 3
EBITS_F6_E3M2, MBITS_F6_E3M2 = 3, 2

SIGN_MASK_F4 = 0x8  # 1000
MANTISSA_MASK_F4 = 0x1  # 0001

ZERO_BITS_F32 = 0x0
ZERO_POINT_FIVE_BITS_F32 = 0x3F000000
F4_E2M1_EXP_BIAS = 1
F32_EXP_BIAS = 127

def triton_f4_to_bf16(x: torch.Tensor):
    """
    Input: a tensor of packed fp4 values
    Output: a tensor of bfloat16 values

    Note: this function is only used in testing, so we can test
      the numerical correctness of the cast without the scaling.
    """
    new_shape = (*x.shape[:-1], x.shape[-1] * 2)
    output = torch.empty(*new_shape, device=x.device, dtype=torch.bfloat16)
    assert x.is_contiguous()
    assert x.is_cuda and output.is_cuda
    n_elements_in = x.numel()
    grid = lambda meta: (  # noqa: E731
        triton.cdiv(n_elements_in, meta["BLOCK_SIZE_IN"]),
    )  # noqa: E731,E501
    triton_f4_to_bf16_kernel[grid](
        x,
        output,
        n_elements_in,
        sign_mask_f4=SIGN_MASK_F4,
        mantissa_mask_f4=MANTISSA_MASK_F4,
        mbits_f4_e2m1=MBITS_F4_E2M1,
        ebits_f4_e2m1=EBITS_F4_E2M1,
        f4_e2m1_exp_bias=F4_E2M1_EXP_BIAS,
        mbits_f32=MBITS_F32,
        ebits_f32=EBITS_F32,
        f32_exp_bias=F32_EXP_BIAS,
        zero_bits_f32=ZERO_BITS_F32,
        zero_point_five_bits_f32=ZERO_POINT_FIVE_BITS_F32,
        BLOCK_SIZE_IN=512,
    )
    return output




##################################################################################################################################################


import torch

def test_triton_f4_to_bf16():
    results = {}
    
    # Test case 1
    n_elements_in = 1024
    x = torch.randint(0, 256, (n_elements_in,), dtype=torch.uint8, device='cuda')
    output = triton_f4_to_bf16(x)
    results["test_case_1"] = output

    # Test case 2
    n_elements_in = 2048
    x = torch.randint(0, 256, (n_elements_in,), dtype=torch.uint8, device='cuda')
    output = triton_f4_to_bf16(x)
    results["test_case_2"] = output

    # Test case 3
    n_elements_in = 512
    x = torch.randint(0, 256, (n_elements_in,), dtype=torch.uint8, device='cuda')
    output = triton_f4_to_bf16(x)
    results["test_case_3"] = output

    # Test case 4
    n_elements_in = 256
    x = torch.randint(0, 256, (n_elements_in,), dtype=torch.uint8, device='cuda')
    output = triton_f4_to_bf16(x)
    results["test_case_4"] = output

    return results

result_gold = test_triton_f4_to_bf16()



import sys
import os
import math
import torch
import triton
import triton.language as tl

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

# Correctly import the operator
from TritonBench_v1.pow_scalar_tensor import pow_func_scalar_tensor_wrapper_rank_1
from performance_utils import Performance_Metrics, do_bench_config

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('pow_scalar_tensor', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(12, 30):
            size = 2 ** i
            in0 = torch.rand(size, dtype=torch.float16)
            out0 = torch.empty(size, dtype=torch.float16)
            self.input_tensors.append((in0, out0))

    def to_cuda(self, input_tensor):
        # return input_tensor.cuda()
        in0, out0 = input_tensor
        return (in0.cuda(), out0.cuda())

    def call_op(self, input_tensor):
        # # Create an output tensor with the same size as input_tensor
        # output_tensor = torch.empty_like(input_tensor)
        # # Call the operator
        # return pow_func_scalar_tensor_wrapper_rank_1(2.0, input_tensor, out0=output_tensor)
        in0, out0 = input_tensor
        return pow_func_scalar_tensor_wrapper_rank_1(2.0, in0, out0=out0)

    def get_gbps(self, input_tensor, runtime):

        # total_bytes = 2 * input_tensor.numel() * input_tensor.element_size()
        in0, out0 = input_tensor
        total_bytes = 2 * in0.numel() * in0.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        # FLOPS = input_tensor.numel()  # One operation per element
        in0, out0 = input_tensor
        FLOPS = in0.numel()
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.chunk_delta_fwd import chunk_fwd_h_fn  # Correctly import the function
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('chunk_delta_fwd', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 11):  # Adjust the range as needed for your testing
            B = 2 ** i
            H = 8  # Number of heads, can be adjusted
            T = 128  # Sequence length, can be adjusted
            K = 64  # Key dimension, can be adjusted
            V = 64  # Value dimension, can be adjusted
            BT = 32  # Block size for T, can be adjusted
            k = torch.rand(B, H, T, K, dtype=torch.float16)
            u = torch.rand(B, H, T, V, dtype=torch.float16)
            w = torch.rand(B, H, T, K, dtype=torch.float16)
            initial_state = torch.rand(B, H, K, V, dtype=torch.float16) if i % 2 == 0 else None
            final_state = torch.rand(B, H, K, V, dtype=torch.float16) if i % 2 == 0 else None
            self.input_tensors.append((k, u, w, BT, initial_state, final_state))

    def to_cuda(self, input_tensor):
        k, u, w, BT, initial_state, final_state = input_tensor
        return (k.cuda(), u.cuda(), w.cuda(), BT, 
                initial_state.cuda() if initial_state is not None else None, 
                final_state.cuda() if final_state is not None else None)

    def call_op(self, input_tensor):
        k, u, w, BT, initial_state, final_state = input_tensor
        return chunk_fwd_h_fn(k, w, u, BT, initial_state, final_state)

    def get_gbps(self, input_tensor, runtime):
        k, u, w, BT, initial_state, final_state = input_tensor
        total_bytes = (k.numel() + u.numel() + w.numel()) * k.element_size()
        if initial_state is not None:
            total_bytes += initial_state.numel() * initial_state.element_size()
        if final_state is not None:
            total_bytes += final_state.numel() * final_state.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        k, u, w, BT, initial_state, final_state = input_tensor
        B, H, T, K, V = k.shape[0], k.shape[1], k.shape[2], k.shape[3], u.shape[3]
        FLOPS = 2 * B * H * T * K * V  # Assuming each operation involves 2 FLOPS
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def cross_entropy_fwd_kernel(
    loss_ptr,  # data ptrs
    lse_ptr,
    z_loss_ptr,
    logits_ptr,
    labels_ptr,
    smoothing,
    logit_scale,
    lse_square_scale,
    ignored_index,
    total_classes,
    class_start_idx,  # Useful for tensor parallel when each rank only has a subset of classes
    n_cols,  # shapes
    n_rows,
    logits_row_stride,  # strides
    BLOCK_SIZE: tl.constexpr,
    HAS_SMOOTHING: tl.constexpr,
    SPLIT: tl.constexpr,
):
    row_idx = tl.program_id(0)
    col_block_idx = tl.program_id(1)
    logits_ptr = logits_ptr + row_idx * logits_row_stride.to(tl.int64)
    col_offsets = col_block_idx * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    label_idx = tl.load(labels_ptr + row_idx)
    logits = tl.load(logits_ptr + col_offsets, mask=col_offsets < n_cols, other=-float("inf")).to(
        tl.float32
    ) * logit_scale
    max_logits = tl.max(logits, 0)
    if HAS_SMOOTHING:
        sum_logits = tl.sum(tl.where(col_offsets < n_cols, logits, 0.0), 0)
    lse = tl.log(tl.sum(tl.exp(logits - max_logits), 0)) + max_logits
    tl.store(lse_ptr + col_block_idx * n_rows + row_idx, lse)
    if label_idx == ignored_index:
        loss = 0.0
        z_loss = 0.0
    else:
        label_idx -= class_start_idx
        if label_idx >= col_block_idx * BLOCK_SIZE and label_idx < min(
            n_cols, (col_block_idx + 1) * BLOCK_SIZE
        ):
            logits_label = tl.load(logits_ptr + label_idx) * logit_scale
            if HAS_SMOOTHING:
                loss = (
                    (lse if not SPLIT else 0.0)
                    - smoothing * sum_logits / total_classes
                    - (1 - smoothing) * logits_label
                )
            else:
                loss = (lse if not SPLIT else 0.0) - logits_label
        else:
            if HAS_SMOOTHING:
                loss = smoothing * ((lse if not SPLIT else 0.0) - sum_logits / total_classes)
            else:
                loss = 0.0
        if not SPLIT:
            z_loss = lse_square_scale * lse * lse
            loss += z_loss
        else:
            z_loss = 0.0
    tl.store(loss_ptr + col_block_idx * n_rows + row_idx, loss)
    if not SPLIT:
        tl.store(z_loss_ptr + col_block_idx * n_rows + row_idx, z_loss)

@triton.jit
def cross_entropy_bwd_kernel(
    dlogits_ptr,  # data ptrs
    dloss_ptr,
    logits_ptr,
    lse_ptr,
    labels_ptr,
    smoothing,
    logit_scale,
    lse_square_scale,
    ignored_index,
    total_classes,
    class_start_idx,  # Useful for tensor parallel when each rank only has a subset of classes
    n_cols,  # shapes
    logits_row_stride,  # strides
    dlogits_row_stride,
    dloss_row_stride,
    BLOCK_SIZE: tl.constexpr,
    HAS_SMOOTHING: tl.constexpr,
):
    row_idx = tl.program_id(0)
    col_block_idx = tl.program_id(1)
    logits_ptr = logits_ptr + row_idx * logits_row_stride.to(tl.int64)
    dlogits_ptr = dlogits_ptr + row_idx * dlogits_row_stride.to(tl.int64)
    col_offsets = col_block_idx * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    label_idx = tl.load(labels_ptr + row_idx)
    if label_idx != ignored_index:
        dloss = tl.load(dloss_ptr + row_idx * dloss_row_stride)
    else:
        dloss = 0.0
    logits = tl.load(logits_ptr + col_offsets, mask=col_offsets < n_cols, other=-float("inf")).to(
        tl.float32
    ) * logit_scale
    lse = tl.load(lse_ptr + row_idx)
    probs = tl.exp(logits - lse)
    probs += 2.0 * lse_square_scale * lse * probs
    label_idx -= class_start_idx
    if HAS_SMOOTHING:
        smooth_negative = smoothing / total_classes
        probs = tl.where(col_offsets == label_idx, probs - (1 - smoothing), probs) - smooth_negative
    else:
        probs = tl.where(col_offsets == label_idx, probs - 1.0, probs)
    tl.store(dlogits_ptr + col_offsets, (dloss * logit_scale) * probs, mask=col_offsets < n_cols)

def cross_entropy_fwd(
    logits, labels, smoothing, logit_scale, lse_square_scale, ignored_index, total_classes, class_start_idx, BLOCK_SIZE, HAS_SMOOTHING, SPLIT
):
    n_rows, n_cols = logits.shape
    loss = torch.empty((n_rows, n_cols), dtype=torch.float32, device=logits.device)
    lse = torch.empty((n_rows, n_cols), dtype=torch.float32, device=logits.device)
    z_loss = torch.empty((n_rows, n_cols), dtype=torch.float32, device=logits.device)
    
    grid = (n_rows, (n_cols + BLOCK_SIZE - 1) // BLOCK_SIZE)
    
    # 调用前向内核，传递相关参数
    cross_entropy_fwd_kernel[grid](
        loss, lse, z_loss, logits, labels, smoothing, logit_scale, lse_square_scale, ignored_index, total_classes, class_start_idx, n_cols, n_rows, logits.stride(0), BLOCK_SIZE, HAS_SMOOTHING, SPLIT
    )
    
    # 打印损失、LSE和z_loss，帮助调试
    print(f"Forward loss: {loss}")
    print(f"Forward LSE: {lse}")
    print(f"Forward z_loss: {z_loss}")
    
    return loss, lse, z_loss

def cross_entropy_bwd(
    dloss, logits, lse, labels, smoothing, logit_scale, lse_square_scale, ignored_index, total_classes, class_start_idx, BLOCK_SIZE, HAS_SMOOTHING
):
    n_rows, n_cols = logits.shape
    dlogits = torch.empty_like(logits)
    
    grid = (n_rows, (n_cols + BLOCK_SIZE - 1) // BLOCK_SIZE)
    
    # 调用反向内核，传递相关参数
    cross_entropy_bwd_kernel[grid](
        dlogits, dloss, logits, lse, labels, smoothing, logit_scale, lse_square_scale, ignored_index, total_classes, class_start_idx, n_cols, logits.stride(0), dlogits.stride(0), dloss.stride(0), BLOCK_SIZE, HAS_SMOOTHING
    )
    
    # 打印反向梯度，帮助调试
    print(f"Backward dlogits: {dlogits}")
    
    return dlogits




##################################################################################################################################################


import torch

def test_cross_entropy_kernels():
    # Test parameters
    n_rows = 4  # Number of rows (batch size)
    n_cols = 8  # Number of columns (number of classes)
    BLOCK_SIZE = 4  # Block size for kernel
    smoothing = 0.1  # Label smoothing factor
    logit_scale = 1.0  # Scale for logits
    lse_square_scale = 0.1  # Scaling for LSE square loss
    ignored_index = -1  # Index to ignore in labels
    total_classes = 10  # Total number of classes
    class_start_idx = 0  # Start index for class partitioning

    # Test data
    logits = torch.randn((n_rows, n_cols), dtype=torch.float32, device='cuda')
    labels = torch.randint(0, n_cols, (n_rows,), dtype=torch.int32, device='cuda')
    dloss = torch.randn((n_rows,), dtype=torch.float32, device='cuda')

    results = {}

    # Test without smoothing and without split
    loss, lse, z_loss = cross_entropy_fwd(logits, labels, 0.0, logit_scale, lse_square_scale, ignored_index, total_classes, class_start_idx, BLOCK_SIZE, False, False)
    dlogits = cross_entropy_bwd(dloss, logits, lse, labels, 0.0, logit_scale, lse_square_scale, ignored_index, total_classes, class_start_idx, BLOCK_SIZE, False)
    results['test_case_1'] = (loss, lse, z_loss, dlogits)

    # Test with smoothing and without split
    loss, lse, z_loss = cross_entropy_fwd(logits, labels, smoothing, logit_scale, lse_square_scale, ignored_index, total_classes, class_start_idx, BLOCK_SIZE, True, False)
    dlogits = cross_entropy_bwd(dloss, logits, lse, labels, smoothing, logit_scale, lse_square_scale, ignored_index, total_classes, class_start_idx, BLOCK_SIZE, True)
    results['test_case_2'] = (loss, lse, z_loss, dlogits)

    # Test with smoothing and with split
    loss, lse, z_loss = cross_entropy_fwd(logits, labels, smoothing, logit_scale, lse_square_scale, ignored_index, total_classes, class_start_idx, BLOCK_SIZE, True, True)
    dlogits = cross_entropy_bwd(dloss, logits, lse, labels, smoothing, logit_scale, lse_square_scale, ignored_index, total_classes, class_start_idx, BLOCK_SIZE, True)
    results['test_case_3'] = (loss, lse, z_loss, dlogits)

    return results

# Run the test cases
result_gold = test_cross_entropy_kernels()
# 分支覆盖率为【3/4】



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl
from typing import Tuple

@triton.jit
def fused_recurrent_fwd_kernel(
    q, k, v, beta, o, h0, ht, s_qk_h, s_vo_h, scale, B, H, T, K: tl.constexpr, V: tl.constexpr, 
    BK: tl.constexpr, BV: tl.constexpr, USE_INITIAL_STATE: tl.constexpr, STORE_FINAL_STATE: tl.constexpr, 
    IS_HEADWISE_BETA: tl.constexpr
):
    i_v, i_k, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    p_q = q + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK)
    p_k = k + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK)
    p_v = v + i_bh * s_vo_h + i_v * BV + tl.arange(0, BV)
    if IS_HEADWISE_BETA:
        p_beta = beta + i_bh * s_vo_h + i_v * BV + tl.arange(0, BV)
    else:
        p_beta = beta + i_bh * T
    p_o = o + (i_bh + i_k * B * H) * s_vo_h + i_v * BV + tl.arange(0, BV)

    mask_bk = (i_k * BK + tl.arange(0, BK)) < K
    mask_bv = (i_v * BV + tl.arange(0, BV)) < V
    mask_kv = mask_bk[None, :] & mask_bv[:, None]

    h = tl.zeros([BV, BK], dtype=tl.float32)

    if USE_INITIAL_STATE:
        p_h0 = h0 + i_bh * K * V + (i_k * BK + tl.arange(0, BK)[None, :]) * V + (i_v * BV + tl.arange(0, BV)[:, None])
        h += tl.load(p_h0, mask=mask_kv, other=0).to(tl.float32)

    for _ in range(0, T):
        b_k = tl.load(p_k, mask=mask_bk, other=0).to(tl.float32)
        b_v = tl.load(p_v, mask=mask_bv, other=0).to(tl.float32)
        b_q = tl.load(p_q, mask=mask_bk, other=0).to(tl.float32) * scale
        _v_minus = tl.sum(h * b_k[None, :], axis=1)
        b_v -= _v_minus
        if IS_HEADWISE_BETA:
            b_beta = tl.load(p_beta, mask=mask_bv, other=0).to(tl.float32)
        else:
            b_beta = tl.load(p_beta).to(tl.float32)
        tl.store(p_v, b_v.to(p_v.dtype.element_ty), mask=mask_bv)
        b_v *= b_beta
        h += b_k[None, :] * b_v[:, None]
        _o = h * b_q[None, :]
        _o = tl.sum(_o, axis=1)
        tl.store(p_o, _o.to(p_o.dtype.element_ty), mask=mask_bv)

        p_q += K
        p_k += K
        p_o += V
        p_v += V
        p_beta += V if IS_HEADWISE_BETA else 1

    if STORE_FINAL_STATE:
        p_ht = ht + i_bh * K * V + (i_k * BK + tl.arange(0, BK)[None, :]) * V + (i_v * BV + tl.arange(0, BV)[:, None])
        tl.store(p_ht, h.to(p_ht.dtype.element_ty), mask=mask_kv)

@triton.jit
def fused_recurrent_bwd_kernel(
    q, k, v, beta, dht, dh0, do, dq, dk, dv, dbeta, h0, s_qk_h, s_vo_h, NK, scale, B, H, T, 
    K: tl.constexpr, V: tl.constexpr, BK: tl.constexpr, BV: tl.constexpr, USE_INITIAL_STATE: tl.constexpr, 
    IS_HEADWISE_BETA: tl.constexpr, USE_DH0: tl.constexpr, USE_DHT: tl.constexpr
):
    i_v, i_k, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    mask_bk = i_k * BK + tl.arange(0, BK) < K
    mask_bv = i_v * BV + tl.arange(0, BV) < V

    p_q = q + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK) + (T - 1) * K
    p_k = k + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK) + (T - 1) * K
    p_do = do + i_bh * s_vo_h + i_v * BV + tl.arange(0, BV) + (T - 1) * V
    p_v = v + i_bh * s_vo_h + i_v * BV + tl.arange(0, BV) + (T - 1) * V
    if IS_HEADWISE_BETA:
        p_beta = beta + i_bh * s_vo_h + i_v * BV + tl.arange(0, BV) + (T - 1) * V
    else:
        p_beta = beta + i_bh * T + T - 1

    p_dk = dk + (i_bh + i_v * B * H) * s_qk_h + i_k * BK + tl.arange(0, BK) + (T - 1) * K
    p_dv = dv + (i_bh + i_k * B * H) * s_vo_h + i_v * BV + tl.arange(0, BV) + (T - 1) * V
    if IS_HEADWISE_BETA:
        p_dbeta = dbeta + (i_bh + i_k * B * H + i_v * B * H * NK) * s_vo_h + tl.arange(0, BV) + (T - 1) * V
    else:
        p_dbeta = dbeta + (i_bh + i_v * B * H) * T + T - 1
    d_h = tl.zeros([BK, BV], dtype=tl.float32)

    if USE_DHT:
        p_ht = dht + i_bh * K * V + (i_k * BK + tl.arange(0, BK)[:, None]) * V + (i_v * BV + tl.arange(0, BV)[None, :])
        d_h += tl.load(p_ht, mask=mask_bk[:, None] & mask_bv[None, :], other=0).to(tl.float32)

    for _ in range(T):
        b_q = tl.load(p_q, mask=mask_bk, other=0).to(tl.float32) * scale
        b_k = tl.load(p_k, mask=mask_bk, other=0).to(tl.float32)
        b_v = tl.load(p_v, mask=mask_bv, other=0).to(tl.float32)
        b_do = tl.load(p_do, mask=mask_bv, other=0).to(tl.float32)
        if IS_HEADWISE_BETA:
            b_beta = tl.load(p_beta, mask=mask_bv, other=0).to(tl.float32)
        else:
            b_beta = tl.load(p_beta).to(tl.float32)
        d_h += b_q[:, None] * b_do[None, :]
        d_k = tl.sum(d_h * (b_v * b_beta)[None, :], axis=1)
        d_v = tl.sum(d_h * b_k[:, None], axis=0)

        d_beta = d_v * b_v if IS_HEADWISE_BETA else tl.sum(d_v * b_v)
        d_v = d_v * b_beta

        tl.store(p_dk, d_k.to(p_dk.dtype.element_ty), mask=mask_bk)
        tl.store(p_dv, d_v.to(p_dv.dtype.element_ty), mask=mask_bv)
        if IS_HEADWISE_BETA:
            tl.store(p_dbeta, d_beta.to(p_dbeta.dtype.element_ty), mask=mask_bv)
        else:
            tl.store(p_dbeta, d_beta.to(p_dbeta.dtype.element_ty))

        d_h -= b_k[:, None] * d_v[None, :]

        p_do -= V
        p_q -= K
        p_k -= K
        p_v -= V
        p_dk -= K
        p_dv -= V
        p_dbeta -= V if IS_HEADWISE_BETA else 1
        p_beta -= V if IS_HEADWISE_BETA else 1

    if USE_DH0:
        p_dh0 = dh0 + i_bh * K * V + (i_k * BK + tl.arange(0, BK)[:, None]) * V + (i_v * BV + tl.arange(0, BV)[None, :])
        tl.store(p_dh0, d_h.to(p_dh0.dtype.element_ty), mask=mask_bk[:, None] & mask_bv[None, :])

    tl.debug_barrier()

    h = tl.zeros([BK, BV], dtype=tl.float32)

    p_q = q + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK)
    p_k = k + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK)
    p_v = v + i_bh * s_vo_h + i_v * BV + tl.arange(0, BV)
    if IS_HEADWISE_BETA:
        p_beta = beta + i_bh * s_vo_h + i_v * BV + tl.arange(0, BV)
    else:
        p_beta = beta + i_bh * T
    p_do = do + i_bh * s_vo_h + i_v * BV + tl.arange(0, BV)
    p_dq = dq + (i_bh + i_v * B * H) * s_qk_h + i_k * BK + tl.arange(0, BK)
    p_dv = dv + (i_bh + i_k * B * H) * s_vo_h + i_v * BV + tl.arange(0, BV)
    p_dk = dk + (i_bh + i_v * B * H) * s_qk_h + i_k * BK + tl.arange(0, BK)

    if USE_INITIAL_STATE:
        mask_kv = mask_bk[:, None] & mask_bv[None, :]
        p_h0 = h0 + i_bh * K * V + (i_k * BK + tl.arange(0, BK)[:, None]) * V + (i_v * BV + tl.arange(0, BV)[None, :])
        h += tl.load(p_h0, mask=mask_kv, other=0).to(tl.float32)

    for i in range(0, T):
        d_k = tl.load(p_dk, mask=mask_bk, other=0).to(tl.float32)
        d_v = tl.load(p_dv, mask=mask_bv, other=0).to(tl.float32)
        d_k -= tl.sum(d_v[None, :] * h, axis=1)
        tl.store(p_dk, d_k.to(p_dk.dtype.element_ty), mask=mask_bk)

        b_k = tl.load(p_k, mask=mask_bk, other=0).to(tl.float32)
        b_v = tl.load(p_v, mask=mask_bv, other=0).to(tl.float32)
        b_do = tl.load(p_do, mask=mask_bv, other=0).to(tl.float32)
        if IS_HEADWISE_BETA:
            b_beta = tl.load(p_beta, mask=mask_bv, other=0).to(tl.float32)
        else:
            b_beta = tl.load(p_beta).to(tl.float32)
        b_v *= b_beta

        h += b_k[:, None] * b_v[None, :]
        _d_q = h * b_do[None, :]
        d_q = tl.sum(_d_q, axis=1) * scale
        tl.store(p_dq, d_q.to(p_dq.dtype.element_ty), mask=mask_bk)

        p_k += K
        p_do += V
        p_v += V
        p_dk += K
        p_dv += V
        p_dq += K
        p_beta += V if IS_HEADWISE_BETA else 1

class FusedRecurrentFunction(torch.autograd.Function):

    @staticmethod
    def forward(ctx, q, k, v, beta, scale=None, initial_state=None, output_final_state=False):
        B, H, T, K, V = *q.shape, v.shape[-1]

        BK, BV = triton.next_power_of_2(K), min(triton.next_power_of_2(V), 8)
        NK, NV = triton.cdiv(K, BK), triton.cdiv(V, BV)
        num_stages = 1
        num_warps = 1
        assert NK == 1, "NK > 1 is not supported yet"
        o = q.new_empty(NK, B, H, T, V)

        if output_final_state:
            final_state = q.new_empty(B, H, K, V, dtype=torch.float32)
        else:
            final_state = None

        grid = (NV, NK, B * H)
        fused_recurrent_fwd_kernel[grid](
            q, k, v, beta, o, initial_state, final_state,
            q.stride(1),
            v.stride(1),
            scale,
            B=B, H=H, T=T, K=K, V=V,
            BK=BK, BV=BV,
            USE_INITIAL_STATE=initial_state is not None,
            STORE_FINAL_STATE=final_state is not None,
            IS_HEADWISE_BETA=beta.ndim == v.ndim,
            num_warps=num_warps,
            num_stages=num_stages,
        )
        o = o.squeeze(0)
        ctx.save_for_backward(q, k, v, beta, initial_state)
        ctx.scale = scale
        return o, final_state

    @staticmethod
    def backward(ctx, do, dht):
        q, k, v, beta, initial_state = ctx.saved_tensors
        B, H, T, K, V = *q.shape, v.shape[-1]
        scale = ctx.scale
        BK, BV = triton.next_power_of_2(K), min(triton.next_power_of_2(V), 32)
        NK, NV = triton.cdiv(K, BK), triton.cdiv(V, BV)
        assert NK == 1, "NK > 1 is not supported yet"
        num_stages = 1
        num_warps = 2

        beta_vector = beta.ndim == v.ndim

        dq = q.new_empty(NV, B, H, T, K)
        dk = q.new_empty(NV, B, H, T, K)
        dv = q.new_empty(NK, B, H, T, V)
        if beta_vector:
            dbeta = q.new_empty(NV, NK, B, H, T, V)
        else:
            dbeta = q.new_empty(NV, B, H, T)
        grid = (NV, NK, B * H)

        if initial_state is not None and initial_state.requires_grad:
            dh0 = torch.empty_like(initial_state, dtype=torch.float32)
        else:
            dh0 = None

        fused_recurrent_bwd_kernel[grid](
            q, k, v, beta, dht, dh0, do, dq, dk, dv, dbeta, initial_state,
            q.stride(1),
            v.stride(1),
            NK, scale,
            B=B, H=H, T=T, K=K, V=V,
            BK=BK, BV=BV,
            USE_INITIAL_STATE=initial_state is not None,
            USE_DH0=dh0 is not None,
            USE_DHT=dht is not None,
            IS_HEADWISE_BETA=beta_vector,
            num_warps=num_warps,
            num_stages=num_stages
        )
        dq = dq.sum(0)
        dk = dk.sum(0)
        dv = dv.sum(0)
        dbeta = dbeta.sum((0, 1)) if beta_vector else dbeta.sum(0)
        return dq.to(q), dk.to(k), dv.to(v), dbeta.to(beta), None, dh0, None

def fused_recurrent_delta_rule(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    beta: torch.Tensor = None,
    scale: float = None,
    initial_state: torch.Tensor = None,
    output_final_state: bool = False
) -> Tuple[torch.Tensor, torch.Tensor]:
    if scale is None:
        scale = q.shape[-1] ** -0.5
    else:
        assert scale > 0, "scale must be positive"
    if beta is None:
        beta = torch.ones_like(q[..., 0])
    o, final_state = FusedRecurrentFunction.apply(q, k, v, beta, scale, initial_state, output_final_state)
    return o, final_state




##################################################################################################################################################


import torch

def test_fused_recurrent_delta_rule_with_backward():
    # 定义尺寸
    B, H, T, K, V = 2, 4, 8, 16, 32

    # 确保输入张量为叶子张量，且 requires_grad=True
    q = torch.randn(B, H, T, K, dtype=torch.float32, device='cuda', requires_grad=True)
    k = torch.randn(B, H, T, K, dtype=torch.float32, device='cuda', requires_grad=True)
    v = torch.randn(B, H, T, V, dtype=torch.float32, device='cuda', requires_grad=True)
    beta_headwise = torch.randn(B, H, T, V, dtype=torch.float32, device='cuda', requires_grad=True)
    beta_non_headwise = torch.randn(B, H, T, dtype=torch.float32, device='cuda', requires_grad=True)
    initial_state = torch.randn(B, H, K, V, dtype=torch.float32, device='cuda', requires_grad=True)

    # Test 1: Headwise beta, with initial_state and final_state
    o, final_state = fused_recurrent_delta_rule(q, k, v, beta=beta_headwise, scale=0.1, initial_state=initial_state, output_final_state=True)

    loss = o.sum() + final_state.sum()
    loss.backward()

    result_1 = {
        "grad_q": q.grad.norm().item(),
        "grad_k": k.grad.norm().item(),
        "grad_v": v.grad.norm().item(),
        "grad_beta_headwise": beta_headwise.grad.norm().item(),
        "grad_initial_state": initial_state.grad.norm().item()
    }

    q.grad.zero_()
    k.grad.zero_()
    v.grad.zero_()
    beta_headwise.grad.zero_()
    initial_state.grad.zero_()

    # Test 2: Non-headwise beta, with initial_state and final_state
    o, final_state = fused_recurrent_delta_rule(q, k, v, beta=beta_non_headwise, scale=0.1, initial_state=initial_state, output_final_state=True)

    loss = o.sum() + final_state.sum()
    loss.backward()

    result_2 = {
        "grad_q": q.grad.norm().item(),
        "grad_k": k.grad.norm().item(),
        "grad_v": v.grad.norm().item(),
        "grad_beta_non_headwise": beta_non_headwise.grad.norm().item(),
        "grad_initial_state": initial_state.grad.norm().item()
    }

    q.grad.zero_()
    k.grad.zero_()
    v.grad.zero_()
    beta_non_headwise.grad.zero_()
    initial_state.grad.zero_()

    # Test 3: No initial state, with final state
    o, final_state = fused_recurrent_delta_rule(q, k, v, beta=beta_headwise, scale=0.1, initial_state=None, output_final_state=True)

    loss = o.sum() + final_state.sum()
    loss.backward()

    result_3 = {
        "grad_q": q.grad.norm().item(),
        "grad_k": k.grad.norm().item(),
        "grad_v": v.grad.norm().item(),
        "grad_beta_headwise": beta_headwise.grad.norm().item()
    }

    q.grad.zero_()
    k.grad.zero_()
    v.grad.zero_()
    beta_headwise.grad.zero_()

    # Test 4: With initial state, no final state output
    o, _ = fused_recurrent_delta_rule(q, k, v, beta=beta_headwise, scale=0.1, initial_state=initial_state, output_final_state=False)

    loss = o.sum()
    loss.backward()

    result_4 = {
        "grad_q": q.grad.norm().item(),
        "grad_k": k.grad.norm().item(),
        "grad_v": v.grad.norm().item(),
        "grad_beta_headwise": beta_headwise.grad.norm().item(),
        "grad_initial_state": initial_state.grad.norm().item()
    }

    return {
        "test_case_1": result_1,
        "test_case_2": result_2,
        "test_case_3": result_3,
        "test_case_4": result_4
    }

result_gold = test_fused_recurrent_delta_rule_with_backward()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.spinning_lock_reduction import spinning_lock
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('spinning_lock_reduction', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(4, 12):  # Example sizes, adjust as needed
            print(i)
            M = N = 2 ** i
            k = 4  # Example value
            BLOCK_SIZE_M = 128
            BLOCK_SIZE_N = 128
            num_sms = 108  # Example value, adjust based on your GPU
            P = torch.zeros((num_sms * BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=torch.float16)
            C = torch.zeros((M, N), dtype=torch.float16)
            stride_cm = C.stride(0)
            stride_cn = C.stride(1)
            locks = torch.zeros(M, dtype=torch.int32)           
            
            self.input_tensors.append((P, C, locks, num_sms, k, M, N, stride_cm, stride_cn, BLOCK_SIZE_M, BLOCK_SIZE_N))

    def to_cuda(self, input_tensor):
        P, C, locks, num_sms, k, M, N, stride_cm, stride_cn, BLOCK_SIZE_M, BLOCK_SIZE_N = input_tensor
        return (P.cuda(), C.cuda(), locks.cuda(), num_sms, k, M, N, stride_cm, stride_cn, BLOCK_SIZE_M, BLOCK_SIZE_N)

    def call_op(self, input_tensor):
        P, C, locks, num_sms, k, M, N, stride_cm, stride_cn, BLOCK_SIZE_M, BLOCK_SIZE_N = input_tensor
        spinning_lock(P, C, locks, num_sms, k, M, N, stride_cm, stride_cn, BLOCK_SIZE_M, BLOCK_SIZE_N)
        return C

    def get_gbps(self, input_tensor, runtime):
        P, C, locks, num_sms, k, M, N, stride_cm, stride_cn, BLOCK_SIZE_M, BLOCK_SIZE_N = input_tensor
        total_bytes = num_sms / k * 9 * (k - 1) * BLOCK_SIZE_M * BLOCK_SIZE_N * 2 + (num_sms - num_sms / k) * 9 * BLOCK_SIZE_M * BLOCK_SIZE_N * 2 + num_sms * 9 * BLOCK_SIZE_M * BLOCK_SIZE_N * 2
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        P, C, locks, num_sms, k, M, N, stride_cm, stride_cn, BLOCK_SIZE_M, BLOCK_SIZE_N = input_tensor
        # Assuming each element in P contributes to a FLOP
        FLOPS = 9 * num_sms * (k - 1) * BLOCK_SIZE_M * BLOCK_SIZE_N / k  # Example calculation, adjust based on actual computation
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.quantize_copy_kv import destindex_copy_quantize_kv
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('quantize_copy_kv', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(4, 20):  # Adjust the range as needed for your tests
            seq_len = 2 ** i
            head_num = 8  # Example head number, adjust as needed
            head_dim = 64  # Example head dimension, adjust as needed
            K = torch.rand(seq_len, head_num, head_dim, dtype=torch.float16)
            DestLoc = torch.randint(0, seq_len, (seq_len,), dtype=torch.int32)
            Out = torch.empty_like(K, dtype=torch.int8)
            Out_scale = torch.randn((seq_len, head_num, 1), dtype=torch.float16)
            self.input_tensors.append((K, DestLoc, Out, Out_scale))

    def to_cuda(self, input_tensor):
        K, DestLoc, Out, Out_scale = input_tensor
        return (K.cuda(), DestLoc.cuda(), Out.cuda(), Out_scale.cuda())

    def call_op(self, input_tensor):
        K, DestLoc, Out, Out_scale = input_tensor
        destindex_copy_quantize_kv(K, DestLoc, Out, Out_scale)
        return Out, Out_scale

    def get_gbps(self, input_tensor, runtime):
        K, DestLoc, Out, Out_scale = input_tensor
        total_bytes = (K.numel() * K.element_size() + DestLoc.numel() * DestLoc.element_size() +
                       Out.numel() * Out.element_size() + Out_scale.numel() * Out_scale.element_size())
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        K, DestLoc, Out, Out_scale = input_tensor
        # Assuming each element in K involves a quantization operation
        FLOPS = 2 * K.numel()  # One for scaling and one for quantization
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import math
import torch
import triton
import triton.language as tl

@triton.autotune(
    configs=[
        triton.Config({'BLOCK_SIZE_H': 1}),
        triton.Config({'BLOCK_SIZE_H': 2}),
        triton.Config({'BLOCK_SIZE_H': 4}),
        triton.Config({'BLOCK_SIZE_H': 8}),
        triton.Config({'BLOCK_SIZE_H': 16}),
        triton.Config({'BLOCK_SIZE_H': 32}),
        triton.Config({'BLOCK_SIZE_H': 64}),
    ],
    key=['chunk_size', 'nheads'],
)
@triton.jit
def _chunk_cumsum_fwd_kernel(
    # Pointers to matrices
    dt_ptr, A_ptr, dt_bias_ptr, dt_out_ptr, dA_cumsum_ptr,
    # Matrix dimension
    batch, seqlen, nheads, chunk_size,
    dt_min, dt_max,
    # Strides
    stride_dt_batch, stride_dt_seqlen, stride_dt_head,
    stride_A_head,
    stride_dt_bias_head,
    stride_dt_out_batch, stride_dt_out_chunk, stride_dt_out_head, stride_dt_out_csize,
    stride_dA_cs_batch, stride_dA_cs_chunk, stride_dA_cs_head, stride_dA_cs_csize,
    # Meta-parameters
    DT_SOFTPLUS: tl.constexpr,
    HAS_DT_BIAS: tl.constexpr,
    BLOCK_SIZE_H: tl.constexpr, BLOCK_SIZE_CHUNK: tl.constexpr,
):
    # Triton kernel implementation for chunked cumulative sum forward pass
    pid_b = tl.program_id(axis=0)
    pid_c = tl.program_id(axis=1)
    pid_h = tl.program_id(axis=2)
    dt_ptr += pid_b * stride_dt_batch + pid_c * chunk_size * stride_dt_seqlen
    dt_out_ptr += pid_b * stride_dt_out_batch + pid_c * stride_dt_out_chunk
    dA_cumsum_ptr += pid_b * stride_dA_cs_batch + pid_c * stride_dA_cs_chunk

    offs_h = pid_h * BLOCK_SIZE_H + tl.arange(0, BLOCK_SIZE_H)
    offs_c = tl.arange(0, BLOCK_SIZE_CHUNK)
    dt_ptrs = dt_ptr + (offs_h[:, None] * stride_dt_head + offs_c[None, :] * stride_dt_seqlen)
    A_ptrs = A_ptr + offs_h * stride_A_head
    dt_out_ptrs = dt_out_ptr + (offs_h[:, None] * stride_dt_out_head + offs_c[None, :] * stride_dt_out_csize)
    dA_cs_ptrs = dA_cumsum_ptr + (offs_h[:, None] * stride_dA_cs_head + offs_c[None, :] * stride_dA_cs_csize)
    chunk_size_limit = min(chunk_size, seqlen - pid_c * chunk_size)

    dt = tl.load(dt_ptrs, mask=(offs_h[:, None] < nheads) & (offs_c[None, :] < chunk_size_limit), other=0.0).to(tl.float32)
    if HAS_DT_BIAS:
        dt_bias = tl.load(dt_bias_ptr + offs_h * stride_dt_bias_head, mask=offs_h < nheads, other=0.0).to(tl.float32)
        dt += dt_bias[:, None]
    if DT_SOFTPLUS:
        dt = tl.where(dt <= 20.0, tl.log(1 + tl.exp(dt)), dt)
    # As of Triton 2.2.0, tl.clamp is not available yet
    # dt = tl.clamp(dt, dt_min, dt_max)
    dt = tl.minimum(tl.maximum(dt, dt_min), dt_max)
    dt = tl.where((offs_h[:, None] < nheads) & (offs_c[None, :] < chunk_size_limit), dt, 0.0)
    tl.store(dt_out_ptrs, dt, mask=(offs_h[:, None] < nheads) & (offs_c[None, :] < chunk_size))
    A = tl.load(A_ptrs, mask=offs_h < nheads, other=0.0).to(tl.float32)
    dA = dt * A[:, None]
    dA_cs = tl.cumsum(dA, axis=1)
    tl.store(dA_cs_ptrs, dA_cs, mask=(offs_h[:, None] < nheads) & (offs_c[None, :] < chunk_size))

def _chunk_cumsum_fwd(dt, A, chunk_size, dt_bias=None, dt_softplus=False, dt_limit=(0.0, float("inf"))):
    """
    Function to perform the forward cumulative sum operation in chunks.

    Arguments:
    - dt: (batch, seqlen, nheads), the input tensor.
    - A: (nheads,), the scaling factors.
    - chunk_size: The size of each chunk to process at a time.
    - dt_bias: (nheads,), optional, biases for dt if applicable.
    - dt_softplus: Boolean, whether to apply the softplus operation to dt.
    - dt_limit: Tuple, (min, max) limits for clamping dt values.

    Returns:
    - dA_cumsum: Cumulative sum result.
    - dt_out: Modified dt after processing.
    """
    batch, seqlen, nheads = dt.shape
    assert A.shape == (nheads,)
    if dt_bias is not None:
        assert dt_bias.shape == (nheads,)
    nchunks = math.ceil(seqlen / chunk_size)
    dt_out = torch.empty(batch, nheads, nchunks, chunk_size, device=dt.device, dtype=torch.float32)
    dA_cumsum = torch.empty(batch, nheads, nchunks, chunk_size, device=dt.device, dtype=torch.float32)
    grid_chunk_cs = lambda META: (batch, nchunks, triton.cdiv(nheads, META['BLOCK_SIZE_H']))
    with torch.cuda.device(dt.device.index):
        _chunk_cumsum_fwd_kernel[grid_chunk_cs](
            dt, A, dt_bias, dt_out, dA_cumsum,
            int(batch), int(seqlen), int(nheads), int(chunk_size),
            dt_limit[0], dt_limit[1],
            dt.stride(0), dt.stride(1), dt.stride(2),
            A.stride(0),
            dt_bias.stride(0) if dt_bias is not None else 0,
            dt_out.stride(0), dt_out.stride(2), dt_out.stride(1), dt_out.stride(3),
            dA_cumsum.stride(0), dA_cumsum.stride(2), dA_cumsum.stride(1), dA_cumsum.stride(3),
            dt_softplus,
            HAS_DT_BIAS=dt_bias is not None,
            BLOCK_SIZE_CHUNK=triton.next_power_of_2(chunk_size),
        )
    return dA_cumsum, dt_out




##################################################################################################################################################


import torch

def test_chunk_cumsum_fwd():
    # Test case 1: Without dt_bias and without dt_softplus
    dt = torch.rand(2, 10, 4, device='cuda')  # (batch, seqlen, nheads)
    A = torch.rand(4, device='cuda')  # (nheads,)
    chunk_size = 5
    dA_cumsum_1, dt_out_1 = _chunk_cumsum_fwd(dt, A, chunk_size)

    # Test case 2: With dt_bias and without dt_softplus
    dt_bias = torch.rand(4, device='cuda')  # (nheads,)
    dA_cumsum_2, dt_out_2 = _chunk_cumsum_fwd(dt, A, chunk_size, dt_bias=dt_bias)

    # Test case 3: Without dt_bias and with dt_softplus
    dA_cumsum_3, dt_out_3 = _chunk_cumsum_fwd(dt, A, chunk_size, dt_softplus=True)

    # Test case 4: With dt_bias and with dt_softplus
    dA_cumsum_4, dt_out_4 = _chunk_cumsum_fwd(dt, A, chunk_size, dt_bias=dt_bias, dt_softplus=True)

    return {
        "test_case_1": (dA_cumsum_1, dt_out_1),
        "test_case_2": (dA_cumsum_2, dt_out_2),
        "test_case_3": (dA_cumsum_3, dt_out_3),
        "test_case_4": (dA_cumsum_4, dt_out_4),
    }

result_gold = test_chunk_cumsum_fwd()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.iv_dependent_matmul import iv_dependent_matmul_wrapper
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('iv_dependent_matmul', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 32):  # Adjust the range as needed for testing
            M = N = K = 128 * i
            self.input_tensors.append((M, K, N))
    
    def to_cuda(self, input_tensor):
        # M, K, N = input_tensor
        # a = torch.rand((M, K), dtype=torch.float16)
        # b = torch.rand((K, N), dtype=torch.float16)
        # return a.cuda(), b.cuda()
        return input_tensor
        
    def call_op(self, input_tensor):
        M, K, N = input_tensor
        BLOCK_SIZE_M = 128  # Example block size, adjust as needed
        BLOCK_SIZE_N = 128
        BLOCK_SIZE_K = 32
        return iv_dependent_matmul_wrapper(M, K, N, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K, type="pre_load")
    
    def get_gbps(self, input_tensor, runtime):
        M, K, N = input_tensor
        total_bytes = (M * K + K * N + M * N) * 4  # 4 bytes for float16
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        M, K, N = input_tensor
        FLOPS = 2 * M * N * K  # 2 * M * N * K operations for matrix multiplication
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

# Correctly import the matmul function
from TritonBench_v1.matmul_triton1 import matmul
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('matmul_triton1', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(4, 7):
            size = 2 ** i
            M = K = N = size
            A = torch.randn((M, K), dtype=torch.float32)
            B = torch.randn((K, N), dtype=torch.float32)
            self.input_tensors.append((A, B))
    def to_cuda(self, input_tensor):
        x, y = input_tensor
        return (x.cuda(), y.cuda())

    def call_op(self, input_tensor):
        x, y = input_tensor
        return matmul(x, y)

    def get_gbps(self, input_tensor, runtime):
        x, y = input_tensor
        total_bytes = (x.numel() + y.numel() + x.size(0) * y.size(1)) * x.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        x, y = input_tensor
        m, k = x.shape
        _, n = y.shape
        FLOPS = 2 * m * n * k  # 2 * M * N * K operations for matrix multiplication
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.kcache_copy_triton import copy_k_to_blocked_cache
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('kcache_copy_triton', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        for i in range(2, 11):  # Adjust the range for desired sizes
            print(i)
            bsz = 2 ** i
            num_kv_heads = 32
            head_dim = 128
            block_size = 32
            num_blocks = 16
            x = 2
            k = torch.rand((bsz, 1, num_kv_heads, head_dim), dtype=torch.float16)
            k_cache = torch.rand((num_blocks, num_kv_heads, head_dim // x, block_size, x), dtype=torch.float16)
            kv_lengths = torch.randint(0, block_size, (bsz,), dtype=torch.int32)
            block_tables = torch.randint(0, block_size, (bsz, block_size), dtype=torch.int32)
            self.input_tensors.append((k, k_cache, kv_lengths, block_tables))

    def to_cuda(self, input_tensor):
        k, k_cache, kv_lengths, block_tables = input_tensor
        return (k.cuda(), k_cache.cuda(), kv_lengths.cuda(), block_tables.cuda())

    def call_op(self, input_tensor):
        k, k_cache, kv_lengths, block_tables = input_tensor
        copy_k_to_blocked_cache(k, k_cache, kv_lengths, block_tables)

    def get_gbps(self, input_tensor, runtime):
        k, k_cache, _, _ = input_tensor
        total_bytes = k.numel() * k.element_size() + k_cache.numel() * k_cache.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        k, _, _, _ = input_tensor
        FLOPS = k.numel()  # Assuming one operation per element
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import triton
import triton.language as tl
import torch
import torch.nn as nn


next_power_of_2 = triton.next_power_of_2
MAX_FUSED_SIZE : int = 65536

def calculate_settings(n : int) -> (int, int,):
    BLOCK_SIZE : int = next_power_of_2(n)
    if BLOCK_SIZE > MAX_FUSED_SIZE:
        raise RuntimeError(f"Cannot launch Triton kernel since n = {n} exceeds "\
                           f"the maximum CUDA blocksize = {MAX_FUSED_SIZE}.")
    num_warps : int = 4
    if   BLOCK_SIZE >= 32768: num_warps = 32
    elif BLOCK_SIZE >=  8192: num_warps = 16
    elif BLOCK_SIZE >=  2048: num_warps = 8
    return BLOCK_SIZE, num_warps

@triton.jit
def _rms_layernorm_forward(
    Y, Y_row_stride,
    X, X_row_stride,
    W, W_row_stride,
    r, r_row_stride,
    n_cols, eps,
    BLOCK_SIZE: tl.constexpr
):
    """
        Fast RMS Layernorm kernel
        Inspiration from a Triton tutorial:
        https://triton-lang.org/main/getting-started/tutorials/05-layer-norm.html
    """
    row_idx = tl.program_id(0)
    col_offsets = tl.arange(0, BLOCK_SIZE)
    mask = col_offsets < n_cols

    Y += row_idx * Y_row_stride
    X += row_idx * X_row_stride
    r += row_idx * r_row_stride

    X_row = tl.load(X + col_offsets, mask=mask, other=0).to(tl.float32)
    W_row = tl.load(W + col_offsets, mask=mask, other=0)

    row_var = tl.sum(X_row * X_row, axis=0) / n_cols
    inv_var = tl.math.rsqrt(row_var + eps)
    tl.store(r, inv_var)
    normed = X_row * inv_var
    normed = normed.to(W_row.dtype)
    output = normed * W_row
    tl.store(Y + col_offsets, output, mask=mask)

@triton.heuristics({"GEMMA": lambda args: args["GEMMA"],})
@triton.jit
def _rms_layernorm_backward(
    dY, dY_row_stride,
    X, X_row_stride,
    W, W_row_stride,
    r, r_row_stride,
    dW, dW_row_stride,
    n_cols, eps,
    GEMMA: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
):
    """
        Fast RMS Layernorm kernel for the backward pass
        Inspiration from a Triton tutorial:
        https://triton-lang.org/main/getting-started/tutorials/05-layer-norm.html
    """
    row_idx = tl.program_id(0)
    col_offsets = tl.arange(0, BLOCK_SIZE)
    mask = col_offsets < n_cols

    dY += row_idx * dY_row_stride
    X  += row_idx *  X_row_stride
    r  += row_idx *  r_row_stride

    dY_row = tl.load(dY + col_offsets, mask=mask, other=0).to(tl.float32)
    X_row  = tl.load(X  + col_offsets, mask=mask, other=0).to(tl.float32)
    W_row  = tl.load(W  + col_offsets, mask=mask, other=0).to(tl.float32)

    inv_var = tl.load(r).to(tl.float32)
    normed = X_row * inv_var

    if GEMMA: dY_W = dY_row * (W_row + 1.0)
    else:     dY_W = dY_row * W_row

    rowsum_dY_normed = tl.sum(dY_W * normed, axis=0)
    output = inv_var/n_cols * (n_cols*dY_W - normed*rowsum_dY_normed)
    tl.store(dY + col_offsets, output, mask=mask)

@triton.jit
def _gemma_rms_layernorm_forward(
    Y, Y_row_stride,
    X, X_row_stride,
    W, W_row_stride,
    r, r_row_stride,
    n_cols, eps,
    BLOCK_SIZE: tl.constexpr,
):
    row_idx = tl.program_id(0)
    col_offsets = tl.arange(0, BLOCK_SIZE)
    mask = col_offsets < n_cols

    Y += row_idx * Y_row_stride
    X += row_idx * X_row_stride
    r += row_idx * r_row_stride

    X_row = tl.load(X + col_offsets, mask=mask, other=0).to(tl.float32)
    W_row = tl.load(W + col_offsets, mask=mask, other=0).to(tl.float32)

    row_var = tl.sum(X_row * X_row, axis=0) / n_cols
    inv_var = tl.math.rsqrt(row_var + eps)
    tl.store(r, inv_var)
    normed = X_row * inv_var
    output = normed * (W_row + 1.0)

    tl.store(Y + col_offsets, output, mask=mask)

class Fast_RMS_Layernorm(torch.autograd.Function):
    @staticmethod
    def forward(ctx, X, W, eps, gemma=False):
        shape = X.shape
        dim = shape[-1]
        X = X.view(-1, dim)
        n_rows, n_cols = X.shape
        BLOCK_SIZE, num_warps = calculate_settings(n_cols)

        Y = torch.empty((n_rows, n_cols), dtype=X.dtype, device="cuda:0")
        r = torch.empty(n_rows, dtype=torch.float32, device="cuda:0")

        fx = _gemma_rms_layernorm_forward if gemma else _rms_layernorm_forward
        fx[(n_rows,)](
            Y, Y.stride(0),
            X, X.stride(0),
            W, W.stride(0),
            r, r.stride(0),
            n_cols, eps,
            BLOCK_SIZE=BLOCK_SIZE,
            num_warps=num_warps,
        )
        ctx.eps = eps
        ctx.BLOCK_SIZE = BLOCK_SIZE
        ctx.num_warps = num_warps
        ctx.GEMMA = gemma
        ctx.save_for_backward(X, W, r)
        return Y.view(*shape)

    @staticmethod
    def backward(ctx, dY):
        shape = dY.shape
        dim = shape[-1]
        dY = dY.view(-1, dim)
        X, W, r = ctx.saved_tensors
        n_rows, n_cols = dY.shape
        dW = X

        _rms_layernorm_backward[(n_rows,)](
            dY, dY.stride(0),
            X,  X.stride(0),
            W,  W.stride(0),
            r,  r.stride(0),
            dW, dW.stride(0),
            n_cols, ctx.eps,
            GEMMA=ctx.GEMMA,
            BLOCK_SIZE=ctx.BLOCK_SIZE,
            num_warps=ctx.num_warps,
        )
        dX = dY.view(*shape)
        return dX, None, None, None

def fast_rms_layernorm(layernorm, X, gemma=False):
    W = layernorm.weight
    eps = layernorm.variance_epsilon if \
        hasattr(layernorm, "variance_epsilon") \
        else layernorm.eps
    out = Fast_RMS_Layernorm.apply(X, W, eps, gemma)
    return out


# Define a simple layernorm module for testing
class SimpleLayerNorm(nn.Module):
    def __init__(self, normalized_shape, eps=1e-5):
        super(SimpleLayerNorm, self).__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape).cuda())
        self.eps = eps



##################################################################################################################################################


# Test function for the fast_rms_layernorm with backward pass
def test_fast_rms_layernorm_with_backward():
    # Create a random input tensor with gradient tracking
    X = torch.randn(2, 4, 8, device='cuda', dtype=torch.float32, requires_grad=True)

    # Create a layernorm instance
    layernorm = SimpleLayerNorm(normalized_shape=X.shape[-1])

    results = {}

    # Test without GEMMA
    output = fast_rms_layernorm(layernorm, X, gemma=False)
    loss = output.mean()  # Compute a dummy loss

    # Perform backward propagation
    loss.backward()
    results['test_case_1'] = X.grad.clone()

    # Reset gradients for the next test
    X.grad.zero_()

    # Test with GEMMA
    output_gemma = fast_rms_layernorm(layernorm, X, gemma=True)
    loss_gemma = output_gemma.mean()  # Compute a dummy loss

    # Perform backward propagation
    loss_gemma.backward()
    results['test_case_2'] = X.grad.clone()

    return results

# Run the test
result_gold = test_fast_rms_layernorm_with_backward()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.rope_transform import rope_forward
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('rope_transform', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 14):  # Adjust the range as needed for different sizes
            batch_size = 2 ** i
            seq_len = 128  # Example sequence length
            n_q_head = 8  # Example number of query heads
            n_kv_head = 8  # Example number of key/value heads
            head_dim = 64  # Example head dimension

            q = torch.rand(batch_size, seq_len, n_q_head, head_dim, dtype=torch.float16)
            k = torch.rand(batch_size, seq_len, n_kv_head, head_dim, dtype=torch.float16)
            cos = torch.rand(seq_len, head_dim // 2, dtype=torch.float16)
            sin = torch.rand(seq_len, head_dim // 2, dtype=torch.float16)

            self.input_tensors.append((q, k, cos, sin))

    def to_cuda(self, input_tensor):
        q, k, cos, sin = input_tensor
        return q.cuda(), k.cuda(), cos.cuda(), sin.cuda()

    def call_op(self, input_tensor):
        q, k, cos, sin = input_tensor
        return rope_forward(q, k, cos, sin)

    def get_gbps(self, input_tensor, runtime):
        q, k, cos, sin = input_tensor
        total_bytes = (q.numel() + k.numel() + cos.numel() + sin.numel()) * q.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        q, k, cos, sin = input_tensor
        # Assuming each element in q and k undergoes a few operations
        FLOPS = 4 * (q.numel() + k.numel())  # Example: 4 operations per element
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import triton
import triton.language as tl
import torch

@triton.autotune(
    configs=[
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 64, 'GROUP_SIZE_M': 8}, num_stages=3, num_warps=8),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=5, num_warps=2),
        triton.Config({'BLOCK_SIZE_M': 32, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=5, num_warps=2),
    ],
    key=['M', 'N', 'K'],
)
@triton.jit
def matmul_kernel(
    a_ptr, b_ptr, c_ptr, M, N, K,
    stride_am, stride_ak, stride_bk, stride_bn, stride_cm, stride_cn,
    BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_K: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr
):
    pid = tl.program_id(axis=0)
    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
    num_pid_in_group = GROUP_SIZE_M * num_pid_n
    group_id = pid // num_pid_in_group
    first_pid_m = group_id * GROUP_SIZE_M
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
    pid_m = first_pid_m + (pid % group_size_m)
    pid_n = (pid % num_pid_in_group) // group_size_m

    offs_am = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_bn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    offs_k = tl.arange(0, BLOCK_SIZE_K)
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)
    b_ptrs = b_ptr + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)

    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        a = tl.load(a_ptrs, mask=offs_k[None, :] < K - k * BLOCK_SIZE_K, other=0.0)
        b = tl.load(b_ptrs, mask=offs_k[:, None] < K - k * BLOCK_SIZE_K, other=0.0)
        accumulator += tl.dot(a, b)
        a_ptrs += BLOCK_SIZE_K * stride_ak
        b_ptrs += BLOCK_SIZE_K * stride_bk

    c_ptrs = c_ptr + (offs_am[:, None] * stride_cm + offs_bn[None, :] * stride_cn)
    c_mask = (offs_am[:, None] < M) & (offs_bn[None, :] < N)
    tl.store(c_ptrs, accumulator, mask=c_mask)

def triton_matmul(a, b):
    assert a.shape[1] == b.shape[0], "Incompatible dimensions"
    M, K = a.shape
    K, N = b.shape

    c = torch.empty((M, N), device=a.device, dtype=a.dtype)
    grid = lambda META: (triton.cdiv(M, META['BLOCK_SIZE_M']) * triton.cdiv(N, META['BLOCK_SIZE_N']),)
    matmul_kernel[grid](
        a, b, c, M, N, K,
        a.stride(0), a.stride(1),
        b.stride(0), b.stride(1),
        c.stride(0), c.stride(1)
    )
    return c



##################################################################################################################################################


import torch

# Function to compare results of Triton and PyTorch matmul
def test_matmul():
    results = {}
    
    # Test case 1
    M, K, N = 256, 256, 256
    a = torch.randn((M, K), device='cuda', dtype=torch.float32)
    b = torch.randn((K, N), device='cuda', dtype=torch.float32)
    c_triton_1 = triton_matmul(a, b)
    results['test_case_1'] = c_triton_1

    # Test case 2
    M, K, N = 64, 64, 64
    a = torch.randn((M, K), device='cuda', dtype=torch.float32)
    b = torch.randn((K, N), device='cuda', dtype=torch.float32)
    c_triton_2 = triton_matmul(a, b)
    results['test_case_2'] = c_triton_2

    # Test case 3
    M, K, N = 16, 16, 16
    a = torch.randn((M, K), device='cuda', dtype=torch.float32)
    b = torch.randn((K, N), device='cuda', dtype=torch.float32)
    c_triton_3 = triton_matmul(a, b)
    results['test_case_3'] = c_triton_3

    return results

# Run the comparison
result_gold = test_matmul()


# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def masked_add_kernel(grad_ptr,
                      p_ptr,
                      p_mask_ptr,
                      n_elements,
                      alpha,
                      BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    p_mask = tl.load(p_mask_ptr + offsets, mask=mask).to(tl.int1)
    mask = mask & ~p_mask
    p = tl.load(p_ptr + offsets, mask=mask)
    grad = tl.load(grad_ptr + offsets, mask=mask)
    grad += p * alpha
    tl.store(grad_ptr + offsets, grad, mask=mask)

def masked_add(grad: torch.Tensor, p_data: torch.Tensor, p_mask: torch.Tensor, alpha: float = 0):
    '''
    equivalent to
    grad.add_(p.data * (1 - p.mask), alpha=decay)
    '''
    assert grad.is_cuda and p_data.is_cuda and p_mask.is_cuda
    assert (grad.layout, p_data.layout, p_mask.layout) == (torch.strided, torch.strided, torch.strided)
    assert grad.stride() == p_data.stride() == p_mask.stride()
    n_elements = grad.numel()
    grid = lambda meta: (triton.cdiv(n_elements, meta['BLOCK_SIZE']),)
    masked_add_kernel[grid](grad, p_data, p_mask, n_elements, alpha, BLOCK_SIZE=1024)



##################################################################################################################################################


import torch

# 测试代码
def test_masked_add():
    # 设置随机种子以保证结果可复现
    torch.manual_seed(0)
    n = 10000  # 选择较大的张量大小

    # 生成随机张量
    grad = torch.randn(n, device='cuda')
    p_data = torch.randn(n, device='cuda')
    p_mask = torch.randint(0, 2, (n,), device='cuda')  # 生成0或1的掩码

    # Triton版本
    results = {}
    
    # Test case 1
    grad_triton = grad.clone()
    masked_add(grad_triton, p_data, p_mask, alpha=0.5)
    results['test_case_1'] = grad_triton.clone()

    # Test case 2: alpha = 0
    grad_triton = grad.clone()
    masked_add(grad_triton, p_data, p_mask, alpha=0)
    results['test_case_2'] = grad_triton.clone()

    # Test case 3: all mask values are 0
    p_mask_zero = torch.zeros(n, device='cuda', dtype=torch.int32)
    grad_triton = grad.clone()
    masked_add(grad_triton, p_data, p_mask_zero, alpha=0.5)
    results['test_case_3'] = grad_triton.clone()

    # Test case 4: all mask values are 1
    p_mask_one = torch.ones(n, device='cuda', dtype=torch.int32)
    grad_triton = grad.clone()
    masked_add(grad_triton, p_data, p_mask_one, alpha=0.5)
    results['test_case_4'] = grad_triton.clone()

    return results

# 运行测试
result_gold = test_masked_add()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def softmax_kernel(output_ptr, input_ptr, input_row_stride, output_row_stride, n_cols, BLOCK_SIZE: tl.constexpr):
    # The rows of the softmax are independent, so we parallelize across those
    row_idx = tl.program_id(0)
    # The stride represents how much we need to increase the pointer to advance 1 row
    row_start_ptr = input_ptr + row_idx * input_row_stride
    # The block size is the next power of two greater than n_cols, so we can fit each
    # row in a single block
    col_offsets = tl.arange(0, BLOCK_SIZE)
    input_ptrs = row_start_ptr + col_offsets
    # Load the row into SRAM, using a mask since BLOCK_SIZE may be > than n_cols
    row = tl.load(input_ptrs, mask=col_offsets < n_cols, other=-float('inf'))
    # Subtract maximum for numerical stability
    row_minus_max = row - tl.max(row, axis=0)
    # Note that exponentiation in Triton is fast but approximate (i.e., think __expf in CUDA)
    numerator = tl.exp(row_minus_max)
    denominator = tl.sum(numerator, axis=0)
    softmax_output = numerator / denominator
    # Write back output to DRAM
    output_row_start_ptr = output_ptr + row_idx * output_row_stride
    output_ptrs = output_row_start_ptr + col_offsets
    tl.store(output_ptrs, softmax_output, mask=col_offsets < n_cols)

def softmax(x):
    n_rows, n_cols = x.shape
    # The block size is the smallest power of two greater than the number of columns in `x`
    BLOCK_SIZE = triton.next_power_of_2(n_cols)
    # Another trick we can use is to ask the compiler to use more threads per row by
    # increasing the number of warps (`num_warps`) over which each row is distributed.
    # You will see in the next tutorial how to auto-tune this value in a more natural
    # way so you don't have to come up with manual heuristics yourself.
    num_warps = 4
    if BLOCK_SIZE >= 2048:
        num_warps = 8
    if BLOCK_SIZE >= 4096:
        num_warps = 16
    # Allocate output
    y = torch.empty_like(x)
    # Enqueue kernel. The 1D launch grid is simple: we have one kernel instance per row o
    # f the input matrix
    softmax_kernel[(n_rows, )](
        y,
        x,
        x.stride(0),
        y.stride(0),
        n_cols,
        num_warps=num_warps,
        BLOCK_SIZE=BLOCK_SIZE,
    )
    return y



##################################################################################################################################################


import torch

def test_softmax():
    # Define the input tensor
    x = torch.randn(128, 512, device='cuda', dtype=torch.float32)

    # Compute softmax using Triton
    output = softmax(x)

    # Additional test cases to cover all branches
    results = {}

    # Test case 1: n_cols < 2048
    x1 = torch.randn(128, 1024, device='cuda', dtype=torch.float32)
    results['test_case_1'] = softmax(x1)

    # Test case 2: 2048 <= n_cols < 4096
    x2 = torch.randn(128, 2048, device='cuda', dtype=torch.float32)
    results['test_case_2'] = softmax(x2)

    # Test case 3: n_cols >= 4096
    x3 = torch.randn(128, 4096, device='cuda', dtype=torch.float32)
    results['test_case_3'] = softmax(x3)

    # Test case 4: n_cols < 2048 (original test case)
    results['test_case_4'] = output

    return results

result_gold = test_softmax()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import triton
import torch
from triton import language as tl
import math

def calculate_lastdim_num_blocks(input_tensor: torch.Tensor, block_size: int) -> int:
    """
    Calculate the number of blocks for a tensor, assuming we
    stride along the last dimension, and a given block size.

    The corresponding pointer arithmetic looks like this:

    ```python
    block_id = tl.program_id(0)
    striding = tl.arange(0, block_size) * stride
    offset = (striding + (block_size * stride * block_id))
    ```

    This function is used to work out the amount of parallel
    work that needs to be done, given as the total number of
    elements divided by the last dimension stride, and a specified
    block size that will then divvy up the work.

    Parameters
    ----------
    input_tensor : torch.Tensor
        Torch N-d tensor to operate over.

    Returns
    -------
    int
        Number of blocks of work, given a block size.
    """
    # get the stride of the last dimension
    stride = input_tensor.stride(-2)
    numel = input_tensor.numel()
    total_blocks = math.ceil(numel / stride)
    return total_blocks


# This kernel computes the fifth order spherical harmonics projections.
@triton.jit
def fifth_order_fwd(
    coord_ptr: tl.tensor,  # Pointer to input coordinates tensor.
    output_ptr: tl.tensor,  # Pointer to output tensor.
    block_size: tl.constexpr,  # Number of elements in each block.
    coord_numel: tl.constexpr,  # Total number of elements in the coordinates tensor.
    output_numel: tl.constexpr,  # Total number of elements in the output tensor.
    col_offset: tl.constexpr,  # Offset for the output tensor.
    output_stride: tl.constexpr,  # Stride of the output tensor.
):
    coord_stride = 3
    block_id = tl.program_id(0)
    coord_striding = tl.arange(0, block_size) * coord_stride
    coord_row_offset = coord_striding + (block_size * coord_stride * block_id)
    x = tl.load(coord_ptr + coord_row_offset, mask=coord_row_offset < coord_numel)
    y = tl.load(
        coord_ptr + coord_row_offset + 1, mask=coord_row_offset + 1 < coord_numel
    )
    z = tl.load(
        coord_ptr + coord_row_offset + 2, mask=coord_row_offset + 2 < coord_numel
    )
    # -------------------- variable and constant definitions
    CONST000 = 1.73430461568895
    CONST001 = 2.32681380862329
    CONST002 = 1.60565407233314
    CONST003 = 3.21130814466628
    CONST004 = 3.31662479035540
    CONST005 = 6.21867148191637
    CONST006 = 6.21867148191637
    CONST007 = 1.60565407233314
    CONST009 = 11.6340690431164
    CONST010 = 12.8452325786651
    CONST011 = 12.4373429638327
    CONST012 = 12.8452325786651
    CONST013 = 13.8744369255116
    CONST017 = 33.9852909359329
    CONST018 = 7.35803132638072
    CONST020 = -44.1481879582843
    CONST021 = -41.6233107765348
    CONST022 = -29.4321253055229
    CONST023 = -23.2681380862329
    CONST024 = -19.2678488679977
    CONST025 = -19.2678488679977
    CONST026 = -16.9926454679664
    CONST027 = -16.9926454679664
    CONST028 = -13.8744369255116
    CONST029 = -16.5831239517770
    CONST030 = 3.46860923137790
    CONST031 = -8.49632273398321
    CONST032 = -5.20291384706685
    CONST033 = -3.46860923137790
    CONST034 = -1.73430461568895
    VAR05 = x * x * x * x * x
    VAR06 = x * x * x * x
    VAR07 = x * x * x
    VAR08 = x * x
    VAR14 = y * y * y * y * y
    VAR15 = y * y * y * y
    VAR16 = y * y * y
    VAR17 = y * y
    VAR23 = z * z * z * z * z
    VAR24 = z * z * z * z
    VAR25 = z * z * z
    VAR26 = z * z
    # -------------------- kernel implementations
    Y00 = CONST001 * VAR05 + CONST009 * VAR24 * x + CONST023 * VAR07 * VAR26
    Y01 = y * (CONST022 * VAR07 * z - CONST022 * VAR25 * x)
    Y02 = (
        CONST000 * VAR05
        + VAR07 * (CONST028 * VAR17 + CONST033 * VAR26)
        + x * (-CONST021 * VAR17 * VAR26 + CONST032 * VAR24)
    )
    Y03 = CONST027 * VAR07 * y * z + x * (CONST017 * VAR16 * z + CONST026 * VAR25 * y)
    Y04 = (
        CONST002 * VAR05
        + VAR07 * (CONST003 * VAR26 + CONST025 * VAR17)
        + x * (CONST002 * VAR24 + CONST010 * VAR15 + CONST024 * VAR17 * VAR26)
    )
    Y05 = (
        CONST004 * VAR14
        + VAR16 * (CONST029 * VAR08 + CONST029 * VAR26)
        + y * (CONST005 * VAR06 + CONST006 * VAR24 + CONST011 * VAR08 * VAR26)
    )
    Y06 = (
        CONST002 * VAR23
        + VAR25 * (CONST003 * VAR08 + CONST024 * VAR17)
        + z * (CONST007 * VAR06 + CONST012 * VAR15 + CONST024 * VAR08 * VAR17)
    )
    Y07 = VAR16 * (CONST026 * VAR08 - CONST026 * VAR26) + y * (
        -CONST031 * VAR06 + CONST031 * VAR24
    )
    Y08 = (
        CONST034 * VAR23
        + VAR25 * (CONST013 * VAR17 + CONST030 * VAR08)
        + z * (CONST021 * VAR08 * VAR17 - CONST032 * VAR06)
    )
    Y09 = y * (CONST018 * VAR06 + CONST018 * VAR24 + CONST020 * VAR08 * VAR26)
    Y10 = CONST001 * VAR23 + CONST009 * VAR06 * z + CONST023 * VAR08 * VAR25
    output_striding = tl.arange(0, block_size) * output_stride
    output_row_offset = (
        output_striding + (block_size * output_stride * block_id) + col_offset
    )
    tl.store(output_ptr + output_row_offset, Y00, mask=output_row_offset < output_numel)
    tl.store(
        output_ptr + output_row_offset + 1,
        Y01,
        mask=output_row_offset + 1 < output_numel,
    )
    tl.store(
        output_ptr + output_row_offset + 2,
        Y02,
        mask=output_row_offset + 2 < output_numel,
    )
    tl.store(
        output_ptr + output_row_offset + 3,
        Y03,
        mask=output_row_offset + 3 < output_numel,
    )
    tl.store(
        output_ptr + output_row_offset + 4,
        Y04,
        mask=output_row_offset + 4 < output_numel,
    )
    tl.store(
        output_ptr + output_row_offset + 5,
        Y05,
        mask=output_row_offset + 5 < output_numel,
    )
    tl.store(
        output_ptr + output_row_offset + 6,
        Y06,
        mask=output_row_offset + 6 < output_numel,
    )
    tl.store(
        output_ptr + output_row_offset + 7,
        Y07,
        mask=output_row_offset + 7 < output_numel,
    )
    tl.store(
        output_ptr + output_row_offset + 8,
        Y08,
        mask=output_row_offset + 8 < output_numel,
    )
    tl.store(
        output_ptr + output_row_offset + 9,
        Y09,
        mask=output_row_offset + 9 < output_numel,
    )
    tl.store(
        output_ptr + output_row_offset + 10,
        Y10,
        mask=output_row_offset + 10 < output_numel,
    )


# This kernel computes the gradients of the fifth order spherical harmonics projections.
@triton.jit
def fifth_order_bwd(
    coord_ptr: tl.tensor,  # Pointer to input coordinates tensor.
    coord_grad_ptr: tl.tensor,  # Pointer to the gradient output tensor.
    sph_grad_ptr: tl.tensor,  # Pointer to the gradient of the spherical harmonic.
    block_size: tl.constexpr,  # Number of elements in each block.
    coord_numel: tl.constexpr,  # Total number of elements in the coordinates tensor.
    output_numel: tl.constexpr,  # Total number of elements in the output tensor.
    col_offset: tl.constexpr,  # Offset for the output tensor.
    output_stride: tl.constexpr,  # Stride of the output tensor.
):
    block_id = tl.program_id(0)
    coord_stride = 3
    coord_striding = tl.arange(0, block_size) * coord_stride
    coord_row_offset = coord_striding + (block_size * coord_stride * block_id)
    x = tl.load(coord_ptr + coord_row_offset, mask=coord_row_offset < coord_numel)
    y = tl.load(
        coord_ptr + coord_row_offset + 1, mask=coord_row_offset + 1 < coord_numel
    )
    z = tl.load(
        coord_ptr + coord_row_offset + 2, mask=coord_row_offset + 2 < coord_numel
    )
    output_striding = tl.arange(0, block_size) * output_stride
    output_row_offset = (
        output_striding + (block_size * output_stride * block_id) + col_offset
    )
    g_0 = tl.load(
        sph_grad_ptr + output_row_offset, mask=output_row_offset < output_numel
    )
    g_1 = tl.load(
        sph_grad_ptr + output_row_offset + 1, mask=output_row_offset + 1 < output_numel
    )
    g_2 = tl.load(
        sph_grad_ptr + output_row_offset + 2, mask=output_row_offset + 2 < output_numel
    )
    g_3 = tl.load(
        sph_grad_ptr + output_row_offset + 3, mask=output_row_offset + 3 < output_numel
    )
    g_4 = tl.load(
        sph_grad_ptr + output_row_offset + 4, mask=output_row_offset + 4 < output_numel
    )
    g_5 = tl.load(
        sph_grad_ptr + output_row_offset + 5, mask=output_row_offset + 5 < output_numel
    )
    g_6 = tl.load(
        sph_grad_ptr + output_row_offset + 6, mask=output_row_offset + 6 < output_numel
    )
    g_7 = tl.load(
        sph_grad_ptr + output_row_offset + 7, mask=output_row_offset + 7 < output_numel
    )
    g_8 = tl.load(
        sph_grad_ptr + output_row_offset + 8, mask=output_row_offset + 8 < output_numel
    )
    g_9 = tl.load(
        sph_grad_ptr + output_row_offset + 9, mask=output_row_offset + 9 < output_numel
    )
    g_10 = tl.load(
        sph_grad_ptr + output_row_offset + 10,
        mask=output_row_offset + 10 < output_numel,
    )
    # -------------------- variable and constant definitions
    CONST000 = 1.60565407233314
    CONST001 = 3.00000000000000
    CONST002 = 3.21130814466628
    CONST003 = 1.60565407233314
    CONST004 = 6.42261628933256
    CONST005 = 6.42261628933256
    CONST006 = 8.67152307844476
    CONST007 = 8.02827036166571
    CONST008 = 6.93721846275580
    CONST009 = 11.6340690431164
    CONST010 = 12.8452325786651
    CONST011 = 6.21867148191637
    CONST012 = 6.21867148191637
    CONST014 = 12.4373429638327
    CONST017 = 12.8452325786651
    CONST018 = 13.8744369255116
    CONST019 = 24.8746859276655
    CONST020 = 24.8746859276655
    CONST021 = 27.7488738510232
    CONST024 = 29.4321253055229
    CONST027 = 7.35803132638072
    CONST029 = 46.5362761724657
    CONST030 = 51.3809303146605
    CONST031 = 51.3809303146605
    CONST034 = 101.955872807799
    CONST036 = -8.67152307844475
    CONST037 = 3.46860923137790
    CONST038 = -88.2963759165686
    CONST039 = -83.2466215530696
    CONST040 = -69.8044142586986
    CONST041 = -50.9779364038993
    CONST042 = -50.9779364038993
    CONST043 = -46.5362761724657
    CONST044 = -44.1481879582843
    CONST045 = -41.6233107765348
    CONST046 = -38.5356977359954
    CONST047 = -38.5356977359954
    CONST048 = -33.1662479035540
    CONST049 = -33.9852909359329
    CONST050 = 6.42261628933257
    CONST051 = -33.9852909359329
    CONST052 = -29.4321253055229
    CONST053 = -27.7488738510232
    CONST054 = -20.8116553882674
    CONST055 = -19.2678488679977
    CONST056 = -19.2678488679977
    CONST057 = -16.9926454679664
    CONST058 = -16.9926454679664
    CONST059 = -13.8744369255116
    CONST060 = -16.5831239517770
    CONST061 = -8.49632273398321
    CONST062 = -6.93721846275580
    CONST063 = -5.20291384706685
    CONST064 = -3.46860923137790
    VAR06 = x * x * x * x
    VAR07 = x * x * x
    VAR08 = x * x
    VAR15 = y * y * y * y
    VAR16 = y * y * y
    VAR17 = y * y
    VAR24 = z * z * z * z
    VAR25 = z * z * z
    VAR26 = z * z
    # -------------------- kernel implementations
    g_x = tl.load(
        coord_grad_ptr + coord_row_offset, mask=coord_row_offset < coord_numel
    )
    g_y = tl.load(
        coord_grad_ptr + coord_row_offset + 1, mask=coord_row_offset + 1 < coord_numel
    )
    g_z = tl.load(
        coord_grad_ptr + coord_row_offset + 2, mask=coord_row_offset + 2 < coord_numel
    )
    g_x += (
        g_0 * (CONST009 * VAR06 + CONST009 * VAR24 + CONST040 * VAR08 * VAR26)
        + g_1 * y * (CONST038 * VAR08 * z - CONST052 * VAR25)
        + g_10 * (CONST029 * VAR07 * z + CONST043 * VAR25 * x)
        + g_2
        * (
            CONST001 * VAR08 * (CONST059 * VAR17 + CONST064 * VAR26)
            + CONST006 * VAR06
            - CONST045 * VAR17 * VAR26
            + CONST063 * VAR24
        )
        + g_3 * (CONST041 * VAR08 * y * z - CONST049 * VAR16 * z + CONST057 * VAR25 * y)
        + g_4
        * (
            CONST000 * VAR24
            + CONST001 * VAR08 * (CONST002 * VAR26 + CONST055 * VAR17)
            + CONST007 * VAR06
            + CONST010 * VAR15
            + CONST056 * VAR17 * VAR26
        )
        + g_5 * (CONST048 * VAR16 * x + y * (CONST019 * VAR07 + CONST019 * VAR26 * x))
        + g_6 * (CONST005 * VAR25 * x + z * (CONST004 * VAR07 + CONST046 * VAR17 * x))
        + g_7 * (CONST049 * VAR16 * x - CONST051 * VAR07 * y)
        + g_8 * (CONST008 * VAR25 * x + z * (CONST039 * VAR17 * x - CONST054 * VAR07))
        + g_9 * y * (CONST024 * VAR07 + CONST038 * VAR26 * x)
    )
    g_y += (
        g_1 * (CONST052 * VAR07 * z - CONST052 * VAR25 * x)
        + g_2 * (-CONST039 * VAR26 * x * y + CONST053 * VAR07 * y)
        + g_3 * (CONST058 * VAR07 * z + x * (CONST034 * VAR17 * z + CONST057 * VAR25))
        + g_4 * (CONST047 * VAR07 * y + x * (CONST030 * VAR16 + CONST046 * VAR26 * y))
        + g_5
        * (
            CONST001 * VAR17 * (CONST060 * VAR08 + CONST060 * VAR26)
            + CONST011 * VAR06
            + CONST012 * VAR24
            + CONST014 * VAR08 * VAR26
            - CONST060 * VAR15
        )
        + g_6 * (CONST046 * VAR25 * y + z * (CONST031 * VAR16 + CONST046 * VAR08 * y))
        + g_7
        * (
            CONST001 * VAR17 * (CONST057 * VAR08 - CONST057 * VAR26)
            - CONST061 * VAR06
            + CONST061 * VAR24
        )
        + g_8 * (CONST021 * VAR25 * y + CONST039 * VAR08 * y * z)
        + g_9 * (CONST027 * VAR06 + CONST027 * VAR24 + CONST044 * VAR08 * VAR26)
    )
    g_z += (
        g_0 * (CONST029 * VAR25 * x + CONST043 * VAR07 * z)
        + g_1 * y * (-CONST038 * VAR26 * x + CONST052 * VAR07)
        + g_10 * (CONST009 * VAR06 + CONST009 * VAR24 + CONST040 * VAR08 * VAR26)
        + g_2 * (CONST062 * VAR07 * z + x * (-CONST039 * VAR17 * z + CONST054 * VAR25))
        + g_3 * (CONST058 * VAR07 * y + x * (CONST042 * VAR26 * y - CONST049 * VAR16))
        + g_4 * (CONST005 * VAR07 * z + x * (CONST046 * VAR17 * z + CONST050 * VAR25))
        + g_5 * (CONST048 * VAR16 * z + y * (CONST019 * VAR08 * z + CONST020 * VAR25))
        + g_6
        * (
            CONST001 * VAR26 * (CONST002 * VAR08 + CONST056 * VAR17)
            + CONST003 * VAR06
            + CONST007 * VAR24
            + CONST017 * VAR15
            + CONST056 * VAR08 * VAR17
        )
        + g_7 * (-CONST049 * VAR16 * z + CONST051 * VAR25 * y)
        + g_8
        * (
            CONST001 * VAR26 * (CONST018 * VAR17 + CONST037 * VAR08)
            + CONST036 * VAR24
            + CONST045 * VAR08 * VAR17
            - CONST063 * VAR06
        )
        + g_9 * y * (CONST024 * VAR25 + CONST038 * VAR08 * z)
    )
    tl.store(
        coord_grad_ptr + coord_row_offset, g_x, mask=coord_row_offset < coord_numel
    )
    tl.store(
        coord_grad_ptr + coord_row_offset + 1,
        g_y,
        mask=coord_row_offset + 1 < coord_numel,
    )
    tl.store(
        coord_grad_ptr + coord_row_offset + 2,
        g_z,
        mask=coord_row_offset + 2 < coord_numel,
    )


class FifthOrderSphericalHarmonic(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        coords: torch.Tensor,  # Input coordinates tensor.
        output_tensor: torch.Tensor | None = None,  # Output tensor placeholder.
        mask: torch.Tensor | None = None,  # Mask tensor (not used in this implementation).
        block_size: int = 64,  # Block size for Triton kernels.
        col_offset: int = 0,  # Column offset for the output tensor.
    ):
        if not isinstance(output_tensor, torch.Tensor):
            output_tensor = torch.empty(
                (*coords.shape[:-1], 11), dtype=coords.dtype, device=coords.device
            )
        coord_numel = coords.numel()
        output_numel = output_tensor.numel()
        num_blocks = calculate_lastdim_num_blocks(coords, block_size)
        fifth_order_fwd[num_blocks,](
            coords,
            output_tensor,
            block_size,
            coord_numel,
            output_numel,
            col_offset,
            output_tensor.stride(-2),
        )
        ctx.save_for_backward(coords)
        return output_tensor

    @staticmethod
    def backward(
        ctx,
        sph_grad_tensor: torch.Tensor,  # Gradient of the output tensor.
        block_size: int = 64,  # Block size for Triton kernels.
        col_offset: int = 0,  # Column offset for the output tensor.
    ) -> torch.Tensor:
        (coords,) = ctx.saved_tensors
        coord_grad_output = torch.zeros_like(coords)
        num_blocks = calculate_lastdim_num_blocks(coords, block_size)
        fifth_order_bwd[num_blocks,](
            coords,
            coord_grad_output,
            sph_grad_tensor,
            block_size,
            coords.numel(),
            sph_grad_tensor.numel(),
            col_offset,
            sph_grad_tensor.stride(-2),
        )
        return coord_grad_output




##################################################################################################################################################


# Test case to compare Triton implementations
def test_fifth_order_spherical_harmonic():
    # Generate random input coordinates
    coords = torch.randn(128, 3, device='cuda', dtype=torch.float32)

    # Triton implementation
    triton_output = FifthOrderSphericalHarmonic.apply(coords)

    # Test for various block sizes and offsets
    test_cases = {}
    for i, (block_size, col_offset) in enumerate([(64, 0), (32, 1), (128, 2), (256, 3)], start=1):
        # 将关键字参数改为位置参数
        triton_output = FifthOrderSphericalHarmonic.apply(coords, None, None, block_size, col_offset)
        test_cases[f'test_case_{i}'] = triton_output

    return test_cases

# Run the test
result_gold = test_fifth_order_spherical_harmonic()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl
# from .utils import get_lora_op_configs

@triton.jit
def _bgmv_shrink_kernel(
    input_ptr,
    lora_ptr,
    out_ptr,
    N,
    K,
    lora_indices,
    scaling,
    xm_stride,
    xk_stride,
    l0_stride,
    lora_k_stride,
    lora_n_stride,
    cm_stride,
    cn_stride,
    BLOCK_N: tl.constexpr,
    BLOCK_K: tl.constexpr,
    SPLIT_K: tl.constexpr,
):
    pid_sk = tl.program_id(axis=0)
    cur_batch = tl.program_id(axis=1)
    lora_index = tl.load(lora_indices + cur_batch)
    if lora_index == -1:
        return

    offset_n = tl.arange(0, BLOCK_N)
    offset_k = tl.arange(0, BLOCK_K) + pid_sk * BLOCK_K
    a_ptr = input_ptr + cur_batch * xm_stride
    b_ptr = lora_ptr + l0_stride * lora_index
    accumulator = tl.zeros((BLOCK_N, ), dtype=tl.float32)
    for k in range(0, K, BLOCK_K * SPLIT_K):
        current_k = k + offset_k
        current_k_c = tl.max_contiguous(current_k, BLOCK_K)
        tiled_a = tl.load(
            a_ptr + current_k_c,
            mask=current_k < K,
            other=0.0,
        )
        b_ptr_mask = (offset_n[:, None] < N) & (current_k[None, :] < K)

        tiled_b = tl.load(
            b_ptr + offset_n[:, None] * lora_k_stride +
            current_k[None, :] * lora_n_stride,
            mask=b_ptr_mask,
            other=0.0,
        )

        accumulator += tl.sum(tiled_a * tiled_b, 1)
    accumulator *= scaling
    offset_cn = tl.arange(0, BLOCK_N)
    c_ptr = out_ptr + cur_batch * cm_stride + offset_cn * cn_stride
    c_mask = offset_cn < N
    if SPLIT_K == 1:
        tl.store(c_ptr, accumulator, mask=c_mask)
    else:
        tl.atomic_add(c_ptr, accumulator, mask=c_mask)


@torch.inference_mode()
def _bgmv_shrink(
    inputs: torch.Tensor,
    lora_a_weights: torch.Tensor,
    output_tensor: torch.Tensor,
    lora_indices_tensor: torch.Tensor,
    scaling: float = 1.0,
) -> None:
    assert inputs.dtype == lora_a_weights.dtype
    assert inputs.dtype in [torch.float16, torch.bfloat16]
    assert lora_a_weights.dtype in [
        torch.float16,
        torch.bfloat16,
    ]
    assert inputs.size(1) == lora_a_weights.size(-1)
    assert inputs.is_contiguous()

    if lora_a_weights.ndim == 4:
        assert lora_a_weights.size(1) == 1
        lora_a_weights = lora_a_weights.squeeze(dim=1)
    else:
        assert lora_a_weights.ndim == 3
    assert lora_a_weights.is_contiguous()
    assert output_tensor.is_contiguous()

    batches = lora_indices_tensor.size(0)
    N, K = lora_a_weights.shape[-2:]
    BLOCK_N = triton.next_power_of_2(N)
    # config = get_lora_op_configs("bgmv_shrink", batches, K)

    grid = lambda META: (
        META["SPLIT_K"],
        batches,
    )
    _bgmv_shrink_kernel[grid](
        inputs,
        lora_a_weights,
        output_tensor,
        N,
        K,
        lora_indices_tensor,
        scaling,
        inputs.stride(0),
        inputs.stride(1),
        lora_a_weights.stride(0),
        lora_a_weights.stride(1),
        lora_a_weights.stride(2),
        output_tensor.stride(0),
        output_tensor.stride(1),
        BLOCK_N=BLOCK_N,
        BLOCK_K=256,
        SPLIT_K=64,

        # **config,
    )
    return




##################################################################################################################################################


import torch

# Test function for _bgmv_shrink
def test_bgmv_shrink():
    # Define input parameters
    batch_size = 2
    N = 16
    K = 32
    scaling = 1.0

    # Create input tensors
    inputs = torch.randn((batch_size, K), dtype=torch.float16, device='cuda').contiguous()
    lora_a_weights = torch.randn((batch_size, 1, N, K), dtype=torch.float16, device='cuda').contiguous()
    output_tensor = torch.zeros((batch_size, N), dtype=torch.float16, device='cuda').contiguous()
    lora_indices_tensor = torch.tensor([0, 1], dtype=torch.int32, device='cuda')

    # Call the _bgmv_shrink function
    _bgmv_shrink(
        inputs=inputs,
        lora_a_weights=lora_a_weights,
        output_tensor=output_tensor,
        lora_indices_tensor=lora_indices_tensor,
        scaling=scaling
    )

    # Store the result in a dictionary
    results = {
        "test_case_1": output_tensor.clone()
    }

    # Additional test cases to cover more branches
    lora_indices_tensor = torch.tensor([-1, 1], dtype=torch.int32, device='cuda')
    _bgmv_shrink(
        inputs=inputs,
        lora_a_weights=lora_a_weights,
        output_tensor=output_tensor,
        lora_indices_tensor=lora_indices_tensor,
        scaling=scaling
    )
    results["test_case_2"] = output_tensor.clone()

    lora_indices_tensor = torch.tensor([0, -1], dtype=torch.int32, device='cuda')
    _bgmv_shrink(
        inputs=inputs,
        lora_a_weights=lora_a_weights,
        output_tensor=output_tensor,
        lora_indices_tensor=lora_indices_tensor,
        scaling=scaling
    )
    results["test_case_3"] = output_tensor.clone()

    lora_indices_tensor = torch.tensor([-1, -1], dtype=torch.int32, device='cuda')
    _bgmv_shrink(
        inputs=inputs,
        lora_a_weights=lora_a_weights,
        output_tensor=output_tensor,
        lora_indices_tensor=lora_indices_tensor,
        scaling=scaling
    )
    results["test_case_4"] = output_tensor.clone()

    return results

# Run the test
result_gold = test_bgmv_shrink()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.matmul_kernel import matmul
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('matmul_kernel', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(6, 32):  # Adjust the range as needed for different sizes
            size = 128 * i
            a = torch.rand(size, size, dtype=torch.float16)
            b = torch.rand(size, size, dtype=torch.float16)
            c = torch.empty(size, size, dtype=torch.float16)
            self.input_tensors.append((c, a, b, size, size, size))

    def to_cuda(self, input_tensor):
        c, a, b, M, N, K = input_tensor
        return (c.cuda(), a.cuda(), b.cuda(), M, N, K)

    def call_op(self, input_tensor):
        c, a, b, M, N, K = input_tensor
        BLOCK_SIZE_M = 128  # Example block size, adjust as needed
        BLOCK_SIZE_N = 128
        BLOCK_SIZE_K = 64
        matmul(c, a, b, M, N, K, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K)
        return c

    def get_gbps(self, input_tensor, runtime):
        c, a, b, M, N, K = input_tensor
        total_bytes = (a.numel() + b.numel() + c.numel()) * a.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        c, a, b, M, N, K = input_tensor
        FLOPS = 2 * M * N * K
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import triton
import triton.language as tl
import torch

@triton.jit
def relu_kernel(x_ptr, out_ptr, N: tl.constexpr, block_size: tl.constexpr):
    # Get the index of the current thread
    pid = tl.program_id(0)
    block_start = pid * block_size
    offsets = block_start + tl.arange(0, block_size)
    mask = offsets < N

    # Load elements from global memory
    x = tl.load(x_ptr + offsets, mask=mask)

    # Compute linear layer
    result = tl.where(x >= 0, x, 0.0)

    # Write result to global memory
    if pid == 0:
        tl.store(out_ptr + offsets, result, mask=mask)

def relu(x):
    # Prepare output tensor
    out = torch.empty_like(x, dtype=torch.float32, device=x.device)
    N = out.numel()

    BLOCK_SIZE = 1024
    num_blocks = (N + BLOCK_SIZE - 1) // BLOCK_SIZE  # Calculate the number of blocks needed

    # Launch Triton kernel
    grid = lambda meta: (triton.cdiv(N, BLOCK_SIZE), )

    relu_kernel[grid](x, out, N, BLOCK_SIZE)

    return out



##################################################################################################################################################


import torch

def test_relu():
    results = {}
    
    # Test case 1: All negative values
    input_tensor = torch.tensor([-3.0, -1.0, -0.5, -2.0, -5.0], dtype=torch.float32, device='cuda')
    output_tensor = relu(input_tensor)
    results['test_case_1'] = output_tensor

    # Test case 2: All positive values
    input_tensor = torch.tensor([3.0, 1.0, 0.5, 2.0, 5.0], dtype=torch.float32, device='cuda')
    output_tensor = relu(input_tensor)
    results['test_case_2'] = output_tensor

    # Test case 3: Mixed values
    input_tensor = torch.tensor([-3.0, -1.0, 0.0, 2.0, 5.0], dtype=torch.float32, device='cuda')
    output_tensor = relu(input_tensor)
    results['test_case_3'] = output_tensor

    # Test case 4: Zero values
    input_tensor = torch.tensor([0.0, 0.0, 0.0, 0.0, 0.0], dtype=torch.float32, device='cuda')
    output_tensor = relu(input_tensor)
    results['test_case_4'] = output_tensor

    return results

result_gold = test_relu()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import logging

import torch
import triton
import triton.language as tl
from typing import Tuple


Shape = Tuple[int]

def broadcastable(s1: Shape, s2: Shape) -> bool:
    r1 = len(s1)
    if r1 == 0:
        return True
    r2 = len(s2)
    if r2 == 0:
        return True

    s1, s2 = (s1, s2) if r1 >= r2 else (s2, s1)
    r1, r2 = (r1, r2) if r1 >= r2 else (r2, r1)

    d = r1 - r2
    for i in range(r2):
        if s1[d + i] == 1 or s2[i] == 1 or s1[d + i] == s2[i]:
            continue
        return False
    return True


def cfggen():
    configs = [
        triton.Config({"BLOCK_SIZE": bs}, num_warps=w)
        for w in [4, 8, 16] #, 32]
        for bs in [256, 512, 1024, 2048, 4096]
    ]
    return configs



@triton.autotune(configs=cfggen(), key=["n_elements"])
@triton.jit
def masked_select_kernel(
    inp_ptr,
    select_mask_ptr,
    prefix_sum_ptr,
    out_ptr,
    n_elements,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tl.program_id(axis=0)
    offsets = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements

    inp = tl.load(inp_ptr + offsets, mask=mask, other=0.0)
    select_mask = tl.load(select_mask_ptr + offsets, mask=mask, other=0.0).to(tl.int1)
    out_offset = tl.load(prefix_sum_ptr + offsets, mask=mask, other=0.0) - 1

    tl.store(out_ptr + out_offset, inp, mask=(select_mask and mask))


def masked_select(inp, mask):
    logging.debug("GEMS MASKED SELECT")

    inp_shape = tuple(inp.shape)
    mask_shape = tuple(mask.shape)

    assert broadcastable(
        inp_shape, mask_shape
    ), "The shapes of the `mask` and the `input` tensor must be broadcastable"
    inp, mask = torch.broadcast_tensors(inp, mask)

    inp = inp.contiguous()
    mask = mask.contiguous()

    mask_flattened = mask.ravel()

    prefix_sum = mask_flattened.cumsum(axis=0)
    out = torch.empty(prefix_sum[-1].item(), dtype=inp.dtype, device=inp.device)

    n_elements = inp.numel()
    grid = lambda meta: (triton.cdiv(n_elements, meta["BLOCK_SIZE"]),)
    with torch.cuda.device(inp.device):
        masked_select_kernel[grid](inp, mask_flattened, prefix_sum, out, n_elements)
    return out




##################################################################################################################################################


def test_masked_select():
    # Initialize a dictionary to store results
    results = {}

    # Test case 9: Random mask for 2D tensor, float32
    x_random = torch.rand((4, 4), device='cuda', dtype=torch.float32)
    mask_random = torch.randint(0, 2, (4, 4), dtype=torch.bool, device='cuda')
    result_random = masked_select(x_random, mask_random)
    results['test_case_0'] = result_random

    # Test case 3: 3D tensor, float64, mask with all True
    x_3d = torch.rand((2, 3, 4), dtype=torch.float64, device='cuda')
    mask_3d = torch.ones((2, 3, 4), dtype=torch.bool, device='cuda')
    result_3d = masked_select(x_3d, mask_3d)
    results['test_case_1'] = result_3d

    # Test case 4: 4D tensor, int64, mask with all False
    x_4d = torch.randint(0, 100, (2, 2, 2, 2), dtype=torch.int64, device='cuda')
    mask_4d = torch.zeros((2, 2, 2, 2), dtype=torch.bool, device='cuda')
    result_4d = masked_select(x_4d, mask_4d)
    results['test_case_2'] = result_4d


    # Test case 13: Large tensor, float32, random mask
    x_large = torch.rand((512, 1024), device='cuda', dtype=torch.float32)
    mask_large = torch.randint(0, 2, (512, 1024), dtype=torch.bool, device='cuda')
    result_large = masked_select(x_large, mask_large)
    results['test_case_3'] = result_large

    return results

result_gold = test_masked_select()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import os
import subprocess
from multiprocessing import Pool, Lock, Value
from tqdm import tqdm
import argparse

def parser():
    parser = argparse.ArgumentParser(description="Run benchmark scripts on multiple GPUs.")
    parser.add_argument("--root_dir", type=str, default="./tmp", help="Directory containing benchmark scripts.")
    args = parser.parse_args()
    return args

gpu_count = 1

# scripts = sorted([f for f in os.listdir(script_dir) if f.endswith(".py")])
# scripts = [os.path.join(script_dir, script) for script in scripts]
# total_scripts = len(scripts)  

progress = Value('i', 0)
progress_lock = Lock()

def run_script(args):
    gpu_id, script = args

    script_name = os.path.basename(script)
    log_file = os.path.join(log_dir, f"{script_name}.log")
    err_file = os.path.join(log_dir, f"{script_name}.err")

    cmd = f"HIP_VISIBLE_DEVICES={gpu_id} python {script}"
    # print(f"Running: {cmd}")

    with open(log_file, "w") as log, open(err_file, "w") as err:
        process = subprocess.Popen(cmd, shell=True, stdout=log, stderr=err)
        process.wait()

    with progress_lock:
        progress.value += 1
        tqdm.write(f"✅ finished {progress.value}/{total_scripts}: {script_name}")


if __name__ == "__main__":
    args = parser()
    script_dir = os.path.join(args.root_dir, "tmp")
    log_dir = os.path.join(args.root_dir, "logs")

    os.makedirs(log_dir, exist_ok=True)

    scripts = sorted([f for f in os.listdir(script_dir) if f.endswith(".py")])
    scripts = [os.path.join(script_dir, script) for script in scripts]
    total_scripts = len(scripts)  

    assert os.path.exists(script_dir), f"Script directory {script_dir} does not exist."

    with Pool(processes=gpu_count) as pool, tqdm(total=total_scripts, desc="Process", ncols=80) as pbar:
        args_list = [(i % gpu_count, scripts[i]) for i in range(total_scripts)]
        
        for _ in pool.imap(run_script, args_list):
            pbar.update(1)

        pool.close()
        pool.join()
